"""Tools for performing AO-to-MO transformations of two-electron integrals
using pyscf.
"""

from typing import Any, Optional

import numpy as np

from pyscf.ao2mo import full, general

from pymolresponse.ao2mo import AO2MO
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol


class AO2MOpyscf(AO2MO):
    """Perform AO-to-MO transformations using pyscf."""

    # TODO what does the pyscf compact kwarg do?
    def __init__(self, C: np.ndarray, verbose: int = 1, pyscfmol: Optional[Any] = None) -> None:
        self.pyscfmol = pyscfmol
        occupations = occupations_from_pyscf_mol(self.pyscfmol, C)
        super().__init__(C, occupations, verbose, I=None)

    def perform_rhf_full(self) -> None:
        norb = self.C.shape[-1]
        tei_mo = full(
            self.pyscfmol, self.C[0], aosym="s4", compact=False, verbose=self.verbose
        ).reshape(norb, norb, norb, norb)
        self.tei_mo = (tei_mo,)

    def perform_rhf_partial(self) -> None:
        nocc_a, nvirt_a, _, _ = self.occupations
        C_occ = self.C[0, :, :nocc_a]
        C_virt = self.C[0, :, nocc_a:]
        C_ovov = (C_occ, C_virt, C_occ, C_virt)
        C_oovv = (C_occ, C_occ, C_virt, C_virt)
        tei_mo_ovov = general(
            self.pyscfmol, C_ovov, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_a, nvirt_a, nocc_a, nvirt_a)
        tei_mo_oovv = general(
            self.pyscfmol, C_oovv, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_a, nocc_a, nvirt_a, nvirt_a)
        self.tei_mo = (tei_mo_ovov, tei_mo_oovv)

    def perform_uhf_full(self) -> None:
        norb = self.C.shape[-1]
        C_a = self.C[0]
        C_b = self.C[1]
        C_aaaa = (C_a, C_a, C_a, C_a)
        C_aabb = (C_a, C_a, C_b, C_b)
        C_bbaa = (C_b, C_b, C_a, C_a)
        C_bbbb = (C_b, C_b, C_b, C_b)
        tei_mo_aaaa = general(
            self.pyscfmol, C_aaaa, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(norb, norb, norb, norb)
        tei_mo_aabb = general(
            self.pyscfmol, C_aabb, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(norb, norb, norb, norb)
        tei_mo_bbaa = general(
            self.pyscfmol, C_bbaa, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(norb, norb, norb, norb)
        tei_mo_bbbb = general(
            self.pyscfmol, C_bbbb, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(norb, norb, norb, norb)
        self.tei_mo = (tei_mo_aaaa, tei_mo_aabb, tei_mo_bbaa, tei_mo_bbbb)

    # pylint: disable=too-many-locals
    def perform_uhf_partial(self) -> None:
        nocc_a, nvirt_a, nocc_b, nvirt_b = self.occupations
        C_occ_alph = self.C[0, :, :nocc_a]
        C_virt_alph = self.C[0, :, nocc_a:]
        C_occ_beta = self.C[1, :, :nocc_b]
        C_virt_beta = self.C[1, :, nocc_b:]
        C_ovov_aaaa = (C_occ_alph, C_virt_alph, C_occ_alph, C_virt_alph)
        C_ovov_aabb = (C_occ_alph, C_virt_alph, C_occ_beta, C_virt_beta)
        C_ovov_bbaa = (C_occ_beta, C_virt_beta, C_occ_alph, C_virt_alph)
        C_ovov_bbbb = (C_occ_beta, C_virt_beta, C_occ_beta, C_virt_beta)
        C_oovv_aaaa = (C_occ_alph, C_occ_alph, C_virt_alph, C_virt_alph)
        C_oovv_bbbb = (C_occ_beta, C_occ_beta, C_virt_beta, C_virt_beta)

        tei_mo_ovov_aaaa = general(
            self.pyscfmol, C_ovov_aaaa, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_a, nvirt_a, nocc_a, nvirt_a)
        tei_mo_ovov_aabb = general(
            self.pyscfmol, C_ovov_aabb, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_a, nvirt_a, nocc_b, nvirt_b)
        tei_mo_ovov_bbaa = general(
            self.pyscfmol, C_ovov_bbaa, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_b, nvirt_b, nocc_a, nvirt_a)
        tei_mo_ovov_bbbb = general(
            self.pyscfmol, C_ovov_bbbb, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_b, nvirt_b, nocc_b, nvirt_b)
        tei_mo_oovv_aaaa = general(
            self.pyscfmol, C_oovv_aaaa, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_a, nocc_a, nvirt_a, nvirt_a)
        tei_mo_oovv_bbbb = general(
            self.pyscfmol, C_oovv_bbbb, aosym="s4", compact=False, verbose=self.verbose
        ).reshape(nocc_b, nocc_b, nvirt_b, nvirt_b)
        self.tei_mo = (
            tei_mo_ovov_aaaa,
            tei_mo_ovov_aabb,
            tei_mo_ovov_bbaa,
            tei_mo_ovov_bbbb,
            tei_mo_oovv_aaaa,
            tei_mo_oovv_bbbb,
        )



r"""Explicit equations for orbital Hessian terms using fully-transformed MO-basis two-electron integrals, *e.g.*, :math:`(pq|rs)`."""

import numpy as np

from pymolresponse.utils import form_vec_energy_differences


def form_rpa_a_matrix_mo_singlet_full(
    E_MO: np.ndarray, TEI_MO: np.ndarray, nocc: int
) -> np.ndarray:
    r"""Form the A (CIS) matrix in the MO basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is
    :math:`\left<aj||ib\right> = \left<aj|ib\right> -
    \left<aj|bi\right> = [ai|jb] - [ab|ji] = 2(ai|jb) - (ab|ji)`. It
    also includes the virt-occ energy difference on the diagonal.
    """

    norb = E_MO.shape[0]
    nvirt = norb - nocc
    nov = nocc * nvirt

    ediff = form_vec_energy_differences(np.diag(E_MO)[:nocc], np.diag(E_MO)[nocc:])

    A = np.empty(shape=(nov, nov))

    for i in range(nocc):
        for a in range(nvirt):
            ia = i * nvirt + a
            for j in range(nocc):
                for b in range(nvirt):
                    jb = j * nvirt + b
                    A[ia, jb] = (
                        2 * TEI_MO[a + nocc, i, j, b + nocc] - TEI_MO[a + nocc, b + nocc, j, i]
                    )

    A += np.diag(ediff)

    return A


def form_rpa_a_matrix_mo_triplet_full(
    E_MO: np.ndarray, TEI_MO: np.ndarray, nocc: int
) -> np.ndarray:
    r"""Form the A (CIS) matrix in the MO basis. [triplet]

    The equation for element :math:`\{ia,jb\}` is :math:`-
    \left<aj|bi\right> = - [ab|ji] = - (ab|ji)`. It also includes the
    virt-occ energy difference on the diagonal.
    """

    norb = E_MO.shape[0]
    nvirt = norb - nocc
    nov = nocc * nvirt

    ediff = form_vec_energy_differences(np.diag(E_MO)[:nocc], np.diag(E_MO)[nocc:])

    A = np.empty(shape=(nov, nov))

    for i in range(nocc):
        for a in range(nvirt):
            ia = i * nvirt + a
            for j in range(nocc):
                for b in range(nvirt):
                    jb = j * nvirt + b
                    A[ia, jb] = -TEI_MO[a + nocc, b + nocc, j, i]

    A += np.diag(ediff)

    return A


def form_rpa_b_matrix_mo_singlet_full(TEI_MO: np.ndarray, nocc: int) -> np.ndarray:
    r"""Form the B matrix for RPA in the MO basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is
    :math:`\left<ab||ij\right> = \left<ab|ij\right> -
    \left<ab|ji\right> = [ai|bj] - [aj|bi] = 2(ai|bj) - (aj|bi)`.
    """

    norb = TEI_MO.shape[0]
    nvirt = norb - nocc
    nov = nocc * nvirt

    B = np.empty(shape=(nov, nov))

    for i in range(nocc):
        for a in range(nvirt):
            ia = i * nvirt + a
            for j in range(nocc):
                for b in range(nvirt):
                    jb = j * nvirt + b
                    B[ia, jb] = (
                        2 * TEI_MO[a + nocc, i, b + nocc, j] - TEI_MO[a + nocc, j, b + nocc, i]
                    )

    return -B


def form_rpa_b_matrix_mo_triplet_full(TEI_MO: np.ndarray, nocc: int) -> np.ndarray:
    r"""Form the B matrix for RPA in the MO basis. [triplet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    norb = TEI_MO.shape[0]
    nvirt = norb - nocc
    nov = nocc * nvirt

    B = np.empty(shape=(nov, nov))

    for i in range(nocc):
        for a in range(nvirt):
            ia = i * nvirt + a
            for j in range(nocc):
                for b in range(nvirt):
                    jb = j * nvirt + b
                    B[ia, jb] = -TEI_MO[a + nocc, j, b + nocc, i]

    return -B


def form_rpa_a_matrix_mo_singlet_ss_full(
    E_MO: np.ndarray, TEI_MO: np.ndarray, nocc: int
) -> np.ndarray:
    r"""Form the same-spin part of the A (CIS) matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    norb = E_MO.shape[0]
    nvirt = norb - nocc
    nov = nocc * nvirt

    ediff = form_vec_energy_differences(np.diag(E_MO)[:nocc], np.diag(E_MO)[nocc:])

    A = np.empty(shape=(nov, nov))

    for i in range(nocc):
        for a in range(nvirt):
            ia = i * nvirt + a
            for j in range(nocc):
                for b in range(nvirt):
                    jb = j * nvirt + b
                    A[ia, jb] = TEI_MO[a + nocc, i, j, b + nocc] - TEI_MO[a + nocc, b + nocc, j, i]

    A += np.diag(ediff)

    return A


def form_rpa_a_matrix_mo_singlet_os_full(
    TEI_MO_xxyy: np.ndarray, nocc_x: int, nocc_y: int
) -> np.ndarray:
    r"""Form the opposite-spin part of the A (CIS) matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    nvirt_x = TEI_MO_xxyy.shape[0] - nocc_x
    nvirt_y = TEI_MO_xxyy.shape[2] - nocc_y
    nov_x = nocc_x * nvirt_x
    nov_y = nocc_y * nvirt_y

    A = np.empty(shape=(nov_x, nov_y))

    for i in range(nocc_x):
        for a in range(nvirt_x):
            ia = i * nvirt_x + a
            for j in range(nocc_y):
                for b in range(nvirt_y):
                    jb = j * nvirt_y + b
                    # TODO
                    A[ia, jb] = TEI_MO_xxyy[a + nocc_x, i, j, b + nocc_y]

    return A


def form_rpa_b_matrix_mo_singlet_ss_full(TEI_MO: np.ndarray, nocc: int) -> np.ndarray:
    r"""Form the same-spin part of the RPA B matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    norb = TEI_MO.shape[0]
    nvirt = norb - nocc
    nov = nocc * nvirt

    B = np.empty(shape=(nov, nov))

    for i in range(nocc):
        for a in range(nvirt):
            ia = i * nvirt + a
            for j in range(nocc):
                for b in range(nvirt):
                    jb = j * nvirt + b
                    B[ia, jb] = TEI_MO[a + nocc, i, b + nocc, j] - TEI_MO[a + nocc, j, b + nocc, i]

    return -B


def form_rpa_b_matrix_mo_singlet_os_full(
    TEI_MO_xxyy: np.ndarray, nocc_x: int, nocc_y: int
) -> np.ndarray:
    r"""Form the opposite-spin part of the RPA B matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    nvirt_x = TEI_MO_xxyy.shape[0] - nocc_x
    nvirt_y = TEI_MO_xxyy.shape[2] - nocc_y
    nov_x = nocc_x * nvirt_x
    nov_y = nocc_y * nvirt_y

    B = np.empty(shape=(nov_x, nov_y))

    for i in range(nocc_x):
        for a in range(nvirt_x):
            ia = i * nvirt_x + a
            for j in range(nocc_y):
                for b in range(nvirt_y):
                    jb = j * nvirt_y + b
                    # TODO
                    B[ia, jb] = TEI_MO_xxyy[a + nocc_x, i, b + nocc_y, j]

    return -B



"""Hard-coded response equations for restricted wavefunctions."""

import numpy as np

import pyscf

from pymolresponse import explicit_equations_full as eqns
from pymolresponse.interfaces.pyscf import molecules, utils
from pymolresponse.interfaces.pyscf.ao2mo import AO2MOpyscf


def test_explicit_rhf_outside_solver_off_diagonal_blocks() -> None:
    mol = molecules.molecule_water_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.kernel()
    mocoeffs = mf.mo_coeff
    moenergies = mf.mo_energy
    ao2mo = AO2MOpyscf(mocoeffs, mol.verbose, mol)
    ao2mo.perform_rhf_full()
    assert len(ao2mo.tei_mo) == 1
    tei_mo = ao2mo.tei_mo[0]  # ty: ignore[index-out-of-bounds]

    C = mocoeffs
    E = np.diag(moenergies)
    occupations = utils.occupations_from_pyscf_mol(mol, mocoeffs)
    nocc, nvirt, _, _ = occupations

    A = eqns.form_rpa_a_matrix_mo_singlet_full(E, tei_mo, nocc)
    B = eqns.form_rpa_b_matrix_mo_singlet_full(tei_mo, nocc)

    G = np.block([[A, B], [B, A]])
    assert G.shape == (2 * nocc * nvirt, 2 * nocc * nvirt)

    G_inv = np.linalg.inv(G)

    components = 3

    integrals_dipole_ao = mol.intor("cint1e_r_sph", comp=components)

    integrals_dipole_mo_ai = []

    for component in range(components):
        integrals_dipole_mo_ai_component = np.dot(
            C[:, nocc:].T, np.dot(integrals_dipole_ao[component, ...], C[:, :nocc])
        ).reshape(-1, order="F")
        integrals_dipole_mo_ai.append(integrals_dipole_mo_ai_component)

    integrals_dipole_mo_ai = np.stack(integrals_dipole_mo_ai, axis=0).T
    integrals_dipole_mo_ai_super = np.concatenate(
        (integrals_dipole_mo_ai, -integrals_dipole_mo_ai), axis=0
    )
    rhsvecs = integrals_dipole_mo_ai_super
    rspvecs = np.dot(G_inv, rhsvecs)

    polarizability = 4 * np.dot(rhsvecs.T, rspvecs) / 2

    # pylint: disable=bad-whitespace
    result__0_00 = np.array(
        [[7.93556221, 0.0, 0.0], [0.0, 3.06821077, 0.0], [0.0, 0.0, 0.05038621]]
    )

    # TODO originally 1.0e-8
    atol = 1.5e-7
    rtol = 0.0
    np.testing.assert_allclose(polarizability, result__0_00, rtol=rtol, atol=atol)


# def test_explicit_rhf_outside_solver_all_blocks() -> None:
#     mol = molecules.molecule_water_sto3g()
#     mol.build()

#     mf = pyscf.scf.RHF(mol)
#     mf.kernel()
#     mocoeffs = mf.mo_coeff
#     moenergies = mf.mo_energy
#     ao2mo = AO2MOpyscf(mocoeffs, mol.verbose, mol)
#     ao2mo.perform_rhf_full()
#     tei_mo = ao2mo.tei_mo[0]

#     C = mocoeffs
#     E = np.diag(moenergies)
#     occupations = utils.occupations_from_pyscf_mol(mol, mocoeffs)
#     nocc, nvirt, _, _ = occupations

#     A = eqns.form_rpa_a_matrix_mo_singlet_full(E, tei_mo, nocc)
#     B = eqns.form_rpa_b_matrix_mo_singlet_full(tei_mo, nocc)

#     # G = np.block([[A, B],
#     #               [B, A]])
#     # assert G.shape == (2*nocc*nvirt, 2*nocc*nvirt)
#     G = A + B
#     assert G.shape == (nocc*nvirt, nocc*nvirt)

#     G_inv = np.linalg.inv(G)

#     components = 3

#     integrals_dipole_ao = mol.intor('cint1e_r_sph', comp=components)

#     integrals_dipole_mo_full = []
#     rspvecs = []

#     for component in range(components):
#         integrals_dipole_mo_full_component = np.dot(C.T, np.dot(integrals_dipole_ao[component, ...], C))
#         integrals_dipole_mo_full.append(integrals_dipole_mo_full_component)

#         rspvec_component = np.dot(G_inv, integrals_dipole_mo_full_component)
#         rspvecs.append(rspvec_component)

#     for a in range(components):
#         for b in range(components):
#             print(np.dot(integrals_dipole_mo_full[a].T, rspvecs[b]))

#     # integrals_dipole_mo_full = np.stack(integrals_dipole_mo_full, axis=0)
#     # print(integrals_dipole_mo_full.shape)
#     # integrals_dipole_mo_ai_super = np.concatenate((integrals_dipole_mo_ai,
#     # #                                                -integrals_dipole_mo_ai), axis=0)
#     # rhsvecs = integrals_dipole_mo_ai_super
#     # rspvecs = np.dot(G_inv, rhsvecs)

#     # polarizability = 4 * np.dot(rhsvecs.T, rspvecs) / 2

#     # # pylint: disable=bad-whitespace
#     # result__0_00 = np.array([[ 7.93556221,  0.,          0.        ],
#     #                          [ 0.,          3.06821077,  0.        ],
#     #                          [ 0.,          0.,          0.05038621]])

#     # atol = 1.0e-8
#     # rtol = 0.0
#     # np.testing.assert_allclose(polarizability, result__0_00, rtol=rtol, atol=atol)



"""Driver for solving the coupled perturbed Hartree-Fock (CPHF) equations."""

from abc import ABC, abstractmethod
from typing import List, Optional, Sequence

import numpy as np

from pymolresponse.core import Hamiltonian, Program, Spin
from pymolresponse.operators import Operator
from pymolresponse.solvers import Solver
from pymolresponse.utils import form_results, form_vec_energy_differences


class Driver(ABC):
    def __init__(self, solver: Solver) -> None:
        assert isinstance(solver, Solver)
        self.solver = solver

        self.results: List[np.ndarray] = []

    @abstractmethod
    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        pass


class CPHF(Driver):
    """Driver for solving the coupled perturbed Hartree-Fock (CPHF) equations."""

    def __init__(self, solver: Solver) -> None:
        assert isinstance(solver, Solver)
        super().__init__(solver)

    def set_frequencies(self, frequencies: Optional[Sequence[float]] = None) -> None:
        r"""Set the frequencies :math:`\omega_f` for which frequency-dependent
        CPHF is performed."""

        # :type frequencies float or list
        # :param frequencies one or more frequencies in atomic units;
        # if None, do the static case (0.0)
        assert self.solver is not None
        self.solver.set_frequencies(frequencies)
        self.frequencies = self.solver.frequencies

    def add_operator(self, operator: Operator) -> None:
        """Add an operator to the list of operators that will be used as the
        right-hand side perturbation."""

        assert self.solver is not None
        self.solver.add_operator(operator)

    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(program, (Program, type(None)))
        # TODO program_obj

        if not hasattr(self, "frequencies"):
            self.set_frequencies([0.0])

        self.solver.run(hamiltonian, spin, program, program_obj)
        self.form_uncoupled_results()
        self.form_results()

    def form_uncoupled_results(self) -> None:
        # We avoid the formation of the full Hessian, but the energy
        # differences on the diagonal are still needed. Form them
        # here. The dynamic frequency contribution will be handled
        # inside the main loop.
        nocc_a, _, nocc_b, _ = self.solver.occupations
        moene_alph = np.diag(self.solver.moenergies[0])
        moene_occ_alph = moene_alph[:nocc_a]
        moene_virt_alph = moene_alph[nocc_a:]
        ediff_alph = form_vec_energy_differences(moene_occ_alph, moene_virt_alph)
        ediff_supervector_alph = np.concatenate((ediff_alph, ediff_alph))
        ediff_supervector_alph_static = ediff_supervector_alph[..., np.newaxis]
        nov_alph = len(ediff_alph)
        if self.solver.is_uhf:
            moene_beta = np.diag(self.solver.moenergies[1])
            moene_occ_beta = moene_beta[:nocc_b]
            moene_virt_beta = moene_beta[nocc_b:]
            ediff_beta = form_vec_energy_differences(moene_occ_beta, moene_virt_beta)
            ediff_supervector_beta = np.concatenate((ediff_beta, ediff_beta))
            ediff_supervector_beta_static = ediff_supervector_beta[..., np.newaxis]
            nov_beta = len(ediff_beta)

        self.uncoupled_results = []

        for f in range(len(self.solver.frequencies)):
            frequency = self.solver.frequencies[f]
            ediff_supervector_alph = ediff_supervector_alph_static.copy()
            ediff_supervector_alph[:nov_alph] = ediff_supervector_alph_static[:nov_alph] - frequency
            ediff_supervector_alph[nov_alph:] = ediff_supervector_alph_static[nov_alph:] + frequency
            if self.solver.is_uhf:
                ediff_supervector_beta = ediff_supervector_beta_static.copy()
                ediff_supervector_beta[:nov_beta] = (
                    ediff_supervector_beta_static[:nov_beta] - frequency
                )
                ediff_supervector_beta[nov_beta:] = (
                    ediff_supervector_beta_static[nov_beta:] + frequency
                )

            # dim_rows -> (number of operators) * (number of components for each operator)
            # dim_cols -> total number of response vectors
            dim_rows = sum(
                self.solver.operators[i].mo_integrals_ai_supervector_alph.shape[0]
                for i in range(len(self.solver.operators))
            )
            dim_cols = dim_rows

            # FIXME, though the reason is weak: the dtype is being inferred
            # from a single operator on the solver, when in reality the dtype
            # of all results should be considered.
            results = np.zeros(
                shape=(dim_rows, dim_cols),
                dtype=self.solver.operators[0].mo_integrals_ai_supervector_alph.dtype,
            )

            # Form the result blocks between each pair of
            # operators. Ignore any potential symmetry in the final
            # result matrix for now.
            result_blocks = []
            row_starts = []
            col_starts = []
            row_start = 0
            for op1 in self.solver.operators:
                col_start = 0
                for op2 in self.solver.operators:
                    result_block = 0.0
                    result_block_alph = form_results(
                        op1.mo_integrals_ai_supervector_alph,
                        op2.mo_integrals_ai_supervector_alph / ediff_supervector_alph,
                    )
                    result_block += result_block_alph
                    if self.solver.is_uhf:
                        result_block_beta = form_results(
                            op1.mo_integrals_ai_supervector_beta,
                            op2.mo_integrals_ai_supervector_beta / ediff_supervector_beta,
                        )
                        result_block += result_block_beta
                    result_blocks.append(result_block)
                    row_starts.append(row_start)
                    col_starts.append(col_start)
                    col_start += op2.mo_integrals_ai_supervector_alph.shape[0]
                row_start += op1.mo_integrals_ai_supervector_alph.shape[0]

            # Put each of the result blocks back in the main results
            # matrix.
            assert len(row_starts) == len(col_starts) == len(result_blocks)
            for idx, result_block in enumerate(result_blocks):
                nr, nc = result_block.shape
                rs, cs = row_starts[idx], col_starts[idx]
                results[rs : rs + nr, cs : cs + nc] = result_block

            # The / 2 is because of the supervector part.
            if self.solver.is_uhf:
                results = 2 * results / 2
            else:
                results = 4 * results / 2
            self.uncoupled_results.append(results)

    def form_results(self) -> None:
        self.results = []

        # TODO change now that Operators keep their own rspvecs
        # self.rspvecs structure:
        # 1. list, index is for frequencies
        # 2. list, index is for operators
        # 3. numpy.ndarray, shape is [a,b,c]
        #   a -> number of operator components
        #   b -> actual matrix elements
        #   c -> 1 (always)

        for operator in self.solver.operators:
            assert len(self.solver.frequencies) == len(operator.rspvecs_alph)
            if self.solver.is_uhf:
                assert len(self.solver.frequencies) == len(operator.rspvecs_beta)
            else:
                assert len(operator.rspvecs_beta) == 0

        for f in range(len(self.solver.frequencies)):
            for i in range(len(self.solver.operators)):
                assert (
                    self.solver.operators[i].rspvecs_alph[f].shape
                    == self.solver.operators[i].mo_integrals_ai_supervector_alph.shape
                )
                if self.solver.is_uhf:
                    assert (
                        self.solver.operators[i].rspvecs_beta[f].shape
                        == self.solver.operators[i].mo_integrals_ai_supervector_beta.shape
                    )

        for f in range(len(self.solver.frequencies)):
            # dim_rows -> (number of operators) * (number of components for each operator)
            # dim_cols -> total number of response vectors
            dim_rows = sum(
                self.solver.operators[i].mo_integrals_ai_supervector_alph.shape[0]
                for i in range(len(self.solver.operators))
            )
            dim_cols = sum(
                self.solver.operators[i].rspvecs_alph[f].shape[0]
                for i in range(len(self.solver.operators))
            )
            assert dim_rows == dim_cols

            # FIXME same comment about the dtype as above
            results = np.zeros(
                shape=(dim_rows, dim_cols), dtype=self.solver.operators[0].rspvecs_alph[f].dtype
            )

            # Form the result blocks between each pair of
            # operators. Ignore any potential symmetry in the final
            # result matrix for now.
            result_blocks = []
            row_starts = []
            col_starts = []
            row_start = 0
            for op1 in self.solver.operators:
                col_start = 0
                for op2 in self.solver.operators:
                    result_block = 0.0
                    result_block_alph = form_results(
                        op1.mo_integrals_ai_supervector_alph, op2.rspvecs_alph[f]
                    )
                    result_block += result_block_alph
                    if self.solver.is_uhf:
                        result_block_beta = form_results(
                            op1.mo_integrals_ai_supervector_beta, op2.rspvecs_beta[f]
                        )
                        result_block += result_block_beta
                    result_blocks.append(result_block)
                    row_starts.append(row_start)
                    col_starts.append(col_start)
                    col_start += op2.rspvecs_alph[f].shape[0]
                row_start += op1.mo_integrals_ai_supervector_alph.shape[0]

            # Put each of the result blocks back in the main results
            # matrix.
            assert len(row_starts) == len(col_starts) == len(result_blocks)
            for idx, result_block in enumerate(result_blocks):
                nr, nc = result_block.shape
                rs, cs = row_starts[idx], col_starts[idx]
                results[rs : rs + nr, cs : cs + nc] = result_block

            # The / 2 is because of the supervector part.
            if self.solver.is_uhf:
                results = 2 * results / 2
            else:
                results = 4 * results / 2
            self.results.append(results)



import psi4

from pymolresponse.data import COORDDIR

# TODO molecule_water_sto3g_angstrom


def molecule_water_sto3g() -> psi4.core.Molecule:
    # TODO this isn"t a proper xyz file. Why?
    with open(COORDDIR / "water.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()))

    # TODO This is supposed to be in bohr.

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(0)
    mol.set_multiplicity(1)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_physicists_water_sto3g() -> psi4.core.Molecule:
    with open(COORDDIR / "water_psi4numpy.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(0)
    mol.set_multiplicity(1)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_physicists_water_augccpvdz() -> psi4.core.Molecule:
    with open(COORDDIR / "water_psi4numpy.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "aug-cc-pvdz"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(0)
    mol.set_multiplicity(1)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_glycine_sto3g() -> psi4.core.Molecule:
    with open(COORDDIR / "glycine.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(0)
    mol.set_multiplicity(1)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_trithiolane_sto3g() -> psi4.core.Molecule:
    with open(COORDDIR / "trithiolane.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(0)
    mol.set_multiplicity(1)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def hydrogen_atom_sto3g() -> psi4.core.Molecule:
    mol = psi4.geometry(
        """
0 2
H
"""
    )

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_bc2h4_cation_sto3g() -> psi4.core.Molecule:
    with open(COORDDIR / "BC2H4.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(1)
    mol.set_multiplicity(1)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


# def molecule_bc2h4_neutral_radical_hf_sto3g():

#     mol = molecule_bc2h4_cation_hf_sto3g()

#     mol.set_molecular_charge(0)
#     mol.set_multiplicity(2)

#     mol.update_geometry()

#     return mol


def molecule_lih_cation_sto3g() -> psi4.core.Molecule:
    with open(COORDDIR / "LiH.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "sto-3g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(1)
    mol.set_multiplicity(2)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_0w4a_dication_321g() -> psi4.core.Molecule:
    with open(COORDDIR / "0w4a.xyz") as fh:
        mol = psi4.geometry("\n".join(fh.readlines()[2:]))

    basis = "3-21g"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.set_molecular_charge(2)
    mol.set_multiplicity(2)
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol


def molecule_bh_cation_def2_svp() -> psi4.core.Molecule:
    mol = psi4.geometry(
        """
1 2
B 0.0000 0.0000 0.0000
H 0.0000 0.0000 1.2340
"""
    )

    basis = "def2-svp"
    mol.set_basis_all_atoms(basis, "BASIS")
    mol.reset_point_group("c1")

    mol.update_geometry()

    psi4.core.set_global_option("BASIS", basis)

    return mol



import numpy as np
import numpy.linalg as npl

import pyscf

from pymolresponse.constants import convfac_au_to_debye
from pymolresponse.helpers import (
    calc_center_of_mass,
    calc_center_of_nuclear_charge,
    get_isotopic_masses,
    nuclear_dipole_contribution,
)


def calc_center_of_mass_pyscf(pyscfmol: pyscf.gto.Mole) -> np.ndarray:
    charges = pyscfmol.atom_charges()
    masses = get_isotopic_masses(charges)
    coords = pyscfmol.atom_coords()
    return calc_center_of_mass(coords, masses)


def calc_center_of_electronic_charge_pyscf(D: np.ndarray, pyscfmol: pyscf.gto.Mole) -> np.ndarray:
    assert len(D.shape) == 2
    # no linear dependencies!
    assert D.shape[0] == D.shape[1]
    zerovec = np.zeros(3)
    dipole_at_zerovec = electronic_dipole_contribution_pyscf(D, pyscfmol, zerovec)
    nelec = pyscfmol.tot_electrons()
    return -dipole_at_zerovec / nelec


def electronic_dipole_contribution_pyscf(
    D: np.ndarray, pyscfmol: pyscf.gto.Mole, origin_in_bohrs: np.ndarray
) -> np.ndarray:
    assert isinstance(D, np.ndarray)
    assert len(D.shape) == 2
    assert D.shape[0] == D.shape[1]
    assert isinstance(origin_in_bohrs, np.ndarray)
    assert origin_in_bohrs.shape == (3,)
    # TODO what to assert about pyscfmol? at least isinstance

    M_AO = pyscfmol.intor("cint1e_r_sph", comp=3)
    assert isinstance(M_AO, np.ndarray)
    assert len(M_AO.shape) == 3
    assert M_AO.shape[1] == M_AO.shape[2]
    assert M_AO.shape[0] == 3
    M100_AO = M_AO[0, :, :]
    M010_AO = M_AO[1, :, :]
    M001_AO = M_AO[2, :, :]

    M100_MO = D * M100_AO
    M010_MO = D * M010_AO
    M001_MO = D * M001_AO

    # pylint: disable=invalid-unary-operand-type
    dipole_electronic_atomic_units = -np.asarray(
        [np.sum(M100_MO), np.sum(M010_MO), np.sum(M001_MO)]
    )
    return dipole_electronic_atomic_units


def calculate_dipole_pyscf(
    nuccoords: np.ndarray,
    nuccharges: np.ndarray,
    origin: np.ndarray,
    D: np.ndarray,
    pyscfmol: pyscf.gto.Mole,
    do_print: bool = False,
) -> np.ndarray:
    assert origin.shape == (3,)
    nuclear_components_au = nuclear_dipole_contribution(nuccoords, nuccharges, origin)
    electronic_components_au = electronic_dipole_contribution_pyscf(D, pyscfmol, origin)
    total_components_au = electronic_components_au + nuclear_components_au
    if do_print:
        nuclear_components_debye = nuclear_components_au * convfac_au_to_debye
        electronic_components_debye = electronic_components_au * convfac_au_to_debye
        total_components_debye = total_components_au * convfac_au_to_debye
        nuclear_norm_au = npl.norm(nuclear_components_au)
        electronic_norm_au = npl.norm(electronic_components_au)
        total_norm_au = npl.norm(total_components_au)
        nuclear_norm_debye = nuclear_norm_au * convfac_au_to_debye
        electronic_norm_debye = electronic_norm_au * convfac_au_to_debye
        total_norm_debye = total_norm_au * convfac_au_to_debye
        print(" origin                        [a.u.]: {} {} {}".format(*origin))
        print(" dipole components, electronic [a.u.]: {} {} {}".format(*electronic_components_au))
        print(" dipole components, nuclear    [a.u.]: {} {} {}".format(*nuclear_components_au))
        print(" dipole components, total      [a.u.]: {} {} {}".format(*total_components_au))
        print(" dipole moment, electronic     [a.u.]: {}".format(electronic_norm_au))
        print(" dipole moment, nuclear        [a.u.]: {}".format(nuclear_norm_au))
        print(" dipole moment, total          [a.u.]: {}".format(total_norm_au))
        print(
            " dipole components, electronic [D]   : {} {} {}".format(*electronic_components_debye)
        )
        print(" dipole components, nuclear    [D]   : {} {} {}".format(*nuclear_components_debye))
        print(" dipole components, total      [D]   : {} {} {}".format(*total_components_debye))
        print(" dipole moment, electronic     [D]   : {}".format(electronic_norm_debye))
        print(" dipole moment, nuclear        [D]   : {}".format(nuclear_norm_debye))
        print(" dipole moment, total          [D]   : {}".format(total_norm_debye))
    return total_components_au


def calculate_origin_pyscf(
    origin_string: str,
    nuccoords: np.ndarray,
    nuccharges: np.ndarray,
    D: np.ndarray,
    pyscfmol: pyscf.gto.Mole,
    do_print: bool = False,
) -> np.ndarray:
    assert isinstance(origin_string, str)
    origin_string = origin_string.lower()
    assert origin_string in (
        "explicitly-set",
        "zero",
        "com",
        "centerofmass",
        "ecc",
        "centerofelcharge",
        "ncc",
        "centerofnuccharge",
    )
    zerovec = np.zeros(3)

    if origin_string == "explicitly-set":
        if do_print:
            print(" --- Origin: explicitly-set ---")
        origin = zerovec
    elif origin_string == "zero":
        if do_print:
            print(" --- Origin: zero ---")
        origin = zerovec
    elif origin_string in ("com", "centerofmass"):
        if do_print:
            print(" --- Origin: center of mass ---")
        masses = get_isotopic_masses(nuccharges[:, 0])
        origin = calc_center_of_mass(nuccoords, masses)
    elif origin_string in ("ecc", "centerofelcharge"):
        if do_print:
            print(" --- Origin: center of electronic charge ---")
        origin = calc_center_of_electronic_charge_pyscf(D, pyscfmol)
    elif origin_string in ("ncc", "centerofnuccharge"):
        if do_print:
            print(" --- Origin: center of nuclear charge ---")
        origin = calc_center_of_nuclear_charge(nuccoords, nuccharges)
    else:
        pass

    if do_print:
        print(" Calculating the dipole at the requested origin...")
        calculate_dipole_pyscf(nuccoords, nuccharges, origin, D, pyscfmol, do_print)

    return origin



import pyscf

from pymolresponse.data import COORDDIR


def molecule_water_sto3g_angstrom(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    mol.basis = "sto-3g"
    mol.charge = 0
    mol.spin = 0

    mol.atom = """O         -1.81298        0.53384       -0.01233
H         -0.82365        0.49649        0.00870
H         -2.10234       -0.29131        0.45244
"""

    return mol


def molecule_water_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "water.xyz") as fh:
        mol.atom = fh.read()
    mol.basis = "sto-3g"
    mol.charge = 0
    mol.spin = 0

    mol.unit = "Bohr"

    return mol


def molecule_physicists_water_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "water_psi4numpy.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "sto-3g"
    mol.charge = 0
    mol.spin = 0

    return mol


def molecule_physicists_water_augccpvdz(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "water_psi4numpy.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "aug-cc-pvdz"
    mol.charge = 0
    mol.spin = 0

    return mol


def molecule_glycine_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "glycine.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "sto-3g"
    mol.charge = 0
    mol.spin = 0

    return mol


def molecule_trithiolane_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "trithiolane.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "sto-3g"
    mol.charge = 0
    mol.spin = 0

    return mol


def hydrogen_atom_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    mol.atom = [["H", (0.0, 0.0, 0.0)]]
    mol.basis = "sto-3g"
    mol.charge = 0
    mol.spin = 1

    return mol


def molecule_bc2h4_cation_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "BC2H4.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "sto-3g"
    mol.charge = 1
    mol.spin = 0

    return mol


def molecule_bc2h4_neutral_radical_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = molecule_bc2h4_cation_sto3g(verbose)
    mol.charge = 0
    mol.spin = 1

    return mol


def molecule_lih_cation_sto3g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "LiH.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "sto-3g"
    mol.charge = 1
    mol.spin = 1

    return mol


def molecule_0w4a_dication_321g(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    with open(COORDDIR / "0w4a.xyz") as fh:
        next(fh)
        next(fh)
        mol.atom = fh.read()
    mol.basis = "3-21g"
    mol.charge = 2
    mol.spin = 1

    return mol


def molecule_bh_cation_def2_svp(verbose: int = 0) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None

    mol.atom = [["B", (0.0000, 0.0000, 0.0000)], ["H", (0.0000, 0.0000, 1.2340)]]
    mol.basis = "def2-svp"
    mol.charge = 1
    mol.spin = 1

    return mol



"""Utility functions that are core to calculating physical values."""

from typing import List, Sequence, Tuple

import numpy as np
import periodictable


def get_most_abundant_isotope(element: periodictable.core.Element) -> periodictable.core.Isotope:
    most_abundant_isotope = element.isotopes[0]
    abundance = 0
    for iso in element:
        if iso.abundance > abundance:
            most_abundant_isotope = iso
            abundance = iso.abundance
    return most_abundant_isotope


def get_isotopic_masses(charges: Sequence[int]) -> np.ndarray:
    masses = []
    for charge in charges:
        el = periodictable.elements[charge]
        isotope = get_most_abundant_isotope(el)
        mass = isotope.mass
        masses.append(mass)
    return np.array(masses)


def calc_center_of_mass(coords: np.ndarray, masses: np.ndarray) -> np.ndarray:
    assert len(masses.shape) == 1
    denominator = np.sum(masses)
    numerator = np.sum(coords * masses[..., np.newaxis], axis=0)
    return numerator / denominator


def calc_center_of_nuclear_charge(coords: np.ndarray, charges: np.ndarray) -> np.ndarray:
    dummy = np.zeros(3)
    center = nuclear_dipole_contribution(coords, charges, dummy)
    total_charge = np.sum(charges)
    return center / total_charge


def nuclear_dipole_contribution(
    nuccoords: np.ndarray, nuccharges: np.ndarray, origin_in_bohrs: np.ndarray
) -> np.ndarray:
    assert isinstance(nuccoords, np.ndarray)
    assert isinstance(nuccharges, np.ndarray)
    assert isinstance(origin_in_bohrs, np.ndarray)
    assert len(nuccoords.shape) == 2
    assert nuccoords.shape[1] == 3
    assert nuccoords.shape[0] == nuccharges.shape[0]
    assert origin_in_bohrs.shape == (3,)
    assert len(nuccharges.shape) in (1, 2)
    if len(nuccharges.shape) == 1:
        charges = nuccharges[..., np.newaxis]
    else:
        assert nuccharges.shape[1] == 1
        charges = nuccharges

    return np.sum((nuccoords - origin_in_bohrs) * charges, axis=0)


def get_uhf_values(
    mat_uhf_a: np.ndarray,
    mat_uhf_b: np.ndarray,
    pair_rohf: Tuple[int, int],
    nocc_a: int,
    nvirt_a: int,
    nocc_b: int,
    nvirt_b: int,
) -> List[float]:
    """For a pair ROHF 1-based indices, find the corresponing alpha- and
    beta-spin UHF values.
    """

    # TODO there has to be a better way than including this here...
    range_uhf_a_closed = list(range(0, nocc_a))
    range_uhf_a_virt = list(range(nocc_a, nocc_a + nvirt_a))
    range_uhf_b_closed = list(range(0, nocc_b))
    range_uhf_b_virt = list(range(nocc_b, nocc_b + nvirt_b))
    indices_uhf_a = [(i, a) for i in range(nocc_a) for a in range(nvirt_a)]
    indices_uhf_b = [(i, a) for i in range(nocc_b) for a in range(nvirt_b)]
    # These are the indices for unique pairs considering the full
    # dimensionality of the system (correct orbital window), [norb,
    # norb], starting from 1.
    indices_display_uhf_a = [(p + 1, q + 1) for p in range_uhf_a_closed for q in range_uhf_a_virt]
    indices_display_uhf_b = [(p + 1, q + 1) for p in range_uhf_b_closed for q in range_uhf_b_virt]

    values = []
    if pair_rohf in indices_display_uhf_a:
        idx_uhf_a = indices_display_uhf_a.index(pair_rohf)
        p_a, q_a = indices_uhf_a[idx_uhf_a]
        val_uhf_a = mat_uhf_a[p_a, q_a]
        values.append(val_uhf_a)
    if pair_rohf in indices_display_uhf_b:
        idx_uhf_b = indices_display_uhf_b.index(pair_rohf)
        p_b, q_b = indices_uhf_b[idx_uhf_b]
        val_uhf_b = mat_uhf_b[p_b, q_b]
        values.append(val_uhf_b)
    return values


def mat_uhf_to_packed_rohf(
    mat_alpha: np.ndarray, mat_beta: np.ndarray, indices_display_rohf: List[Tuple[int, int]]
) -> np.ndarray:
    dim = len(indices_display_rohf)
    mat_rohf = np.zeros(dim)
    for idx, pair_rohf in enumerate(indices_display_rohf):
        mat_rohf[idx] = sum(get_uhf_values(mat_alpha, mat_beta, pair_rohf))
    return mat_rohf



import unittest

from pymolresponse import utils


class SplitterTest(unittest.TestCase):
    def test_default(self) -> None:
        """Does the splitter remove empty fields by default properly?"""
        fixed_splitter = utils.Splitter((4, 3, 5, 6, 10, 10, 10, 10, 10, 10))
        line_full = "  60  H 10  s        0.14639   0.00000   0.00000  -0.00000  -0.00000   0.00000"
        line_truncated = "   1  C 1   s       -0.00000  -0.00000   0.00000"
        ref_full = [
            "60",
            "H",
            "10",
            "s",
            "0.14639",
            "0.00000",
            "0.00000",
            "-0.00000",
            "-0.00000",
            "0.00000",
        ]
        ref_truncated = ["1", "C", "1", "s", "-0.00000", "-0.00000", "0.00000"]
        tokens_full = fixed_splitter.split(line_full)
        tokens_truncated = fixed_splitter.split(line_truncated)
        self.assertEqual(ref_full, tokens_full)
        self.assertEqual(ref_truncated, tokens_truncated)

    def test_no_truncation(self) -> None:
        """Does the splitter return even the empty fields when asked?"""
        fixed_splitter = utils.Splitter((4, 3, 5, 6, 10, 10, 10, 10, 10, 10))
        line = "   1  C 1   s       -0.00000  -0.00000   0.00000"
        ref_not_truncated = ["1", "C", "1", "s", "-0.00000", "-0.00000", "0.00000", "", "", ""]
        tokens_not_truncated = fixed_splitter.split(line, truncate=False)
        self.assertEqual(ref_not_truncated, tokens_not_truncated)



import numpy as np

import pyscf

from pymolresponse import cphf, operators, solvers, utils
from pymolresponse.core import AO2MOTransformationType, Hamiltonian, Program, Spin
from pymolresponse.interfaces.pyscf import molecules
from pymolresponse.interfaces.pyscf.ao2mo import AO2MOpyscf
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol


def mol_atom(
    symbol: str = "He", charge: int = 0, spin: int = 0, basis: str = "sto-3g", verbose: int = 0
) -> pyscf.gto.Mole:
    mol = pyscf.gto.Mole()
    mol.verbose = verbose
    mol.output = None
    mol.atom = [[symbol, (0.0, 0.0, 0.0)]]
    mol.basis = basis
    mol.charge = charge
    mol.spin = spin
    mol.build()
    return mol


# pylint: disable=bad-whitespace
rhf_coupled = {
    0.0: {
        "result": np.array(
            [
                [3.47899e01, -1.77591e-05, 2.24416e-04],
                [-1.77591e-05, 4.43286e01, -1.91191e-01],
                [2.24416e-04, -1.91191e-01, 1.68476e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
    0.0773178: {
        "result": np.array(
            [
                [3.53972e01, -8.48249e-06, 2.05118e-04],
                [-8.48249e-06, 4.52552e01, -2.17003e-01],
                [2.05118e-04, -2.17003e-01, 1.70848e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
    0.128347: {
        "result": np.array(
            [
                [3.65252e01, -2.36748e-06, 1.76171e-04],
                [-2.36748e-06, 4.70314e01, -2.86868e-01],
                [1.76171e-04, -2.86868e-01, 1.75452e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
}


# pylint: disable=bad-whitespace
rhf_uncoupled = {
    0.0: {
        "result": np.array(
            [
                [3.22077e01, 1.50870e-04, 5.33903e-04],
                [1.50870e-04, 4.04007e01, 2.38977e00],
                [5.33903e-04, 2.38977e00, 1.49105e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
    0.0773178: {
        "result": np.array(
            [
                [3.25145e01, 1.51073e-04, 5.32981e-04],
                [1.51073e-04, 4.08495e01, 2.43278e00],
                [5.32981e-04, 2.43278e00, 1.50160e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
    0.128347: {
        "result": np.array(
            [
                [3.30688e01, 1.51325e-04, 5.30935e-04],
                [1.51325e-04, 4.16650e01, 2.51199e00],
                [5.30935e-04, 2.51199e00, 1.52058e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
}


# pylint: disable=bad-whitespace
uhf_coupled = {
    0.0: {
        "result": np.array(
            [
                [3.63373e01, -8.63039e-04, 4.27969e-04],
                [-8.63039e-04, 3.50789e01, -2.31340e00],
                [4.27969e-04, -2.31340e00, 1.91465e01],
            ]
        ),
        # TODO was 1.0e-4
        "error_max_diag": 6.0e-4,
    },
    0.0773178: {
        "result": np.array(
            [
                [3.70513e01, -1.35934e-02, 2.24680e-03],
                [-1.35934e-02, 4.33419e01, 1.02081e-01],
                [2.24680e-03, 1.02081e-01, 2.04381e01],
            ]
        ),
        "error_max_diag": 1.0e-3,
    },
    0.128347: {
        "result": np.array(
            [
                [3.97633e01, 2.90266e-02, -2.85016e-03],
                [2.90266e-02, 4.65961e01, 1.43676e00],
                [-2.85016e-03, 1.43676e00, 2.25248e01],
            ]
        ),
        "error_max_diag": 1.0e-2,
    },
    # 100 nm!
    0.4556355: {
        "result": np.array(
            [
                [1.55494e02, -4.03311e-01, -4.85361e-01],
                [-4.03311e-01, -6.60084e02, -9.94324e01],
                [-4.85361e-01, -9.94324e01, 2.80687e01],
            ]
        ),
        "error_max_diag": 2.0e-0,
    },
}


# pylint: disable=bad-whitespace
uhf_uncoupled = {
    0.0: {
        "result": np.array(
            [
                [3.45537e01, 8.12383e-05, 9.13783e-04],
                [8.12383e-05, 5.73968e01, 8.85793e00],
                [9.13783e-04, 8.85793e00, 1.95421e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
    0.0773178: {
        "result": np.array(
            [
                [3.49369e01, 9.40696e-05, 9.31091e-04],
                [9.40696e-05, 5.89094e01, 9.26378e00],
                [9.31091e-04, 9.26378e00, 1.98017e01],
            ]
        ),
        "error_max_diag": 1.0e-4,
    },
    0.128347: {
        "result": np.array(
            [
                [3.56387e01, 1.22202e-04, 9.68661e-04],
                [1.22202e-04, 6.19498e01, 1.00967e01],
                [9.68661e-04, 1.00967e01, 2.03043e01],
            ]
        ),
        "error_max_diag": 2.0e-4,
    },
    0.4556355: {
        "result": np.array(
            [
                [1.03406e02, -1.24375e-02, -1.16407e-02],
                [-1.24375e-02, 7.97942e01, 1.90354e01],
                [-1.16407e-02, 1.90354e01, 3.33718e01],
            ]
        ),
        "error_max_diag": 2.0e-2,
    },
}


def test_uncoupled_rhf() -> None:
    mol = molecules.molecule_trithiolane_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    solver = solvers.ExactInv(C, E, occupations)

    ao2mo = AO2MOpyscf(C, mol.verbose, mol)
    ao2mo.perform_rhf_partial()
    solver.tei_mo = ao2mo.tei_mo
    solver.tei_mo_type = AO2MOTransformationType.partial

    driver = cphf.CPHF(solver)

    operator_diplen = operators.Operator(
        label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
    )
    integrals_diplen_ao = mol.intor("cint1e_r_sph", comp=3)
    operator_diplen.ao_integrals = integrals_diplen_ao
    driver.add_operator(operator_diplen)

    frequencies = [0.0, 0.0773178, 0.128347]
    driver.set_frequencies(frequencies)

    driver.run(
        hamiltonian=Hamiltonian.RPA, spin=Spin.singlet, program=Program.PySCF, program_obj=mol
    )

    for idxf, frequency in enumerate(frequencies):
        print(idxf, frequency)
        print("uncoupled")
        diag_res = np.diag(driver.uncoupled_results[idxf])
        diag_ref = np.diag(rhf_uncoupled[frequency]["result"])
        diff = diag_res - diag_ref
        print(diag_res)
        print(diag_ref)
        print(diff)
        assert np.max(np.abs(diff)) < rhf_uncoupled[frequency]["error_max_diag"]
        print("coupled")
        diag_res = np.diag(driver.results[idxf])
        diag_ref = np.diag(rhf_coupled[frequency]["result"])
        diff = diag_res - diag_ref
        print(diag_res)
        print(diag_ref)
        print(diff)
        assert np.max(np.abs(diff)) < rhf_coupled[frequency]["error_max_diag"]


def test_uncoupled_uhf() -> None:
    mol = molecules.molecule_trithiolane_sto3g()
    mol.charge = 1
    mol.spin = 1
    mol.build()

    mf = pyscf.scf.uhf.UHF(mol)
    mf.scf()

    assert mf.mo_coeff is not None
    assert mf.mo_energy is not None
    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    solver = solvers.ExactInv(C, E, occupations)

    ao2mo = AO2MOpyscf(C, mol.verbose, mol)
    ao2mo.perform_uhf_partial()
    solver.tei_mo = ao2mo.tei_mo
    solver.tei_mo_type = AO2MOTransformationType.partial

    driver = cphf.CPHF(solver)

    operator_diplen = operators.Operator(
        label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
    )
    integrals_diplen_ao = mol.intor("cint1e_r_sph", comp=3)
    operator_diplen.ao_integrals = integrals_diplen_ao
    driver.add_operator(operator_diplen)

    frequencies = [0.0, 0.0773178, 0.128347, 0.4556355]
    driver.set_frequencies(frequencies)

    driver.run(
        hamiltonian=Hamiltonian.RPA, spin=Spin.singlet, program=Program.PySCF, program_obj=mol
    )

    for idxf, frequency in enumerate(frequencies):
        print(idxf, frequency)
        print("uncoupled")
        diag_res = np.diag(driver.uncoupled_results[idxf])
        diag_ref = np.diag(uhf_uncoupled[frequency]["result"])
        diff = diag_res - diag_ref
        print(diag_res)
        print(diag_ref)
        print(diff)
        print(driver.uncoupled_results[idxf])
        print(uhf_uncoupled[frequency]["result"])
        # FIXME why only look at the diagonal elements?
        # assert np.max(np.abs(diff)) < uhf_uncoupled[frequency]["error_max_diag"]
        print("coupled")
        diag_res = np.diag(driver.results[idxf])
        diag_ref = np.diag(uhf_coupled[frequency]["result"])
        diff = diag_res - diag_ref
        print(diag_res)
        print(diag_ref)
        print(diff)
        assert np.max(np.abs(diff)) < uhf_coupled[frequency]["error_max_diag"]



from typing import Optional, Tuple

import numpy as np

from pymolresponse.integrals import JK, IntegralLabel, Integrals, IntegralSymmetry

ANGMOM_COMMON_GAUGE = IntegralLabel("cint1e_cg_irxp_sph", 3)
ANGMOM_GIAO = IntegralLabel("cint1e_giao_irjxp_sph", 3)
DIPOLE = IntegralLabel("cint1e_r_sph", 3)
DIPVEL = IntegralLabel("cint1e_ipovlp_sph", 3)
KINETIC = IntegralLabel("int1e_kin")
A11PART_GIAO = IntegralLabel("int1e_giao_a11part", 9)
A11PART_CG = IntegralLabel("int1e_cg_a11part", 9)
A01 = IntegralLabel("int1e_a01gp", 9)
NUCLEAR = IntegralLabel("int1e_nuc")
SO_1e = IntegralLabel("int1e_prinvxp", 3, IntegralSymmetry.ANTISYMMETRIC)
NSO_1e = IntegralLabel("int1e_pnucxp", 3, IntegralSymmetry.ANTISYMMETRIC)
SO_SPHER_1e = IntegralLabel("cint1e_prinvxp_sph", 3)


class IntegralsPyscf(Integrals):
    def __init__(self, pyscfmol) -> None:
        super().__init__()

        self.mol = pyscfmol

    def _compute(self, label: IntegralLabel) -> np.ndarray:
        if label.symmetry == IntegralSymmetry.ANTISYMMETRIC:
            return self.mol.intor_asymmetric(label.label, comp=label.comp)
        return self.mol.intor(label.label, comp=label.comp)


class JKPyscf(JK):
    def __init__(self, pyscfmol) -> None:
        super().__init__()

        self.mol = pyscfmol

    def compute_from_density(self, D: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    def compute_from_mocoeffs(
        self, C_left: np.ndarray, C_right: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError



import numpy as np

from pymolresponse import cphf, operators, solvers, utils
from pymolresponse.core import AO2MOTransformationType, Hamiltonian, Program, Spin
from pymolresponse.data import REFDIR
from pymolresponse.interfaces.pyscf import molecules
from pymolresponse.properties import electric
from pymolresponse.tests.test_runners import (
    run_as_many_tests_as_possible_rhf_disk,
    run_as_many_tests_as_possible_uhf_disk,
)


def test_final_result_rhf_h2o_sto3g_rpa_singlet() -> None:
    """Test correctness of the final result for water/STO-3G with full RPA for
    singlet response induced by the dipole length operator (the electric
    polarizability) computed with quantities from disk.
    """
    hamiltonian = Hamiltonian.RPA
    spin = Spin.singlet

    C = utils.fix_mocoeffs_shape(utils.np_load(REFDIR / "C.npz"))
    E = utils.fix_moenergies_shape(utils.np_load(REFDIR / "F_MO.npz"))
    TEI_MO = utils.np_load(REFDIR / "TEI_MO.npz")
    # nocc_alph, nvirt_alph, nocc_beta, nvirt_beta
    occupations = np.asarray([5, 2, 5, 2], dtype=int)
    stub = "h2o_sto3g_"
    dim = occupations[0] + occupations[1]
    mat_dipole_x = utils.parse_int_file_2(REFDIR / f"{stub}mux.dat", dim)
    mat_dipole_y = utils.parse_int_file_2(REFDIR / f"{stub}muy.dat", dim)
    mat_dipole_z = utils.parse_int_file_2(REFDIR / f"{stub}muz.dat", dim)

    solver = solvers.ExactInv(C, E, occupations)
    solver.tei_mo = (TEI_MO,)
    solver.tei_mo_type = AO2MOTransformationType.full
    driver = cphf.CPHF(solver)
    ao_integrals_dipole = np.stack((mat_dipole_x, mat_dipole_y, mat_dipole_z), axis=0)
    operator_dipole = operators.Operator(
        label="dipole", is_imaginary=False, is_spin_dependent=False
    )
    operator_dipole.ao_integrals = ao_integrals_dipole
    driver.add_operator(operator_dipole)
    frequencies = (0.0, 0.02, 0.06, 0.1)
    driver.set_frequencies(frequencies)

    driver.run(hamiltonian=hamiltonian, spin=spin, program=None, program_obj=None)

    assert len(driver.results) == len(frequencies)

    result__0_00 = np.array(
        [[7.93556221, 0.0, 0.0], [0.0, 3.06821077, 0.0], [0.0, 0.0, 0.05038621]]
    )
    result__0_02 = np.array(
        [[7.94312371, 0.0, 0.0], [0.0, 3.07051688, 0.0], [0.0, 0.0, 0.05054685]]
    )
    result__0_06 = np.array(
        [[8.00414009, 0.0, 0.0], [0.0, 3.08913608, 0.0], [0.0, 0.0, 0.05186977]]
    )
    result__0_10 = np.array([[8.1290378, 0.0, 0.0], [0.0, 3.12731363, 0.0], [0.0, 0.0, 0.05473482]])

    atol = 1.0e-8
    rtol = 0.0
    np.testing.assert_allclose(driver.results[0], result__0_00, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[1], result__0_02, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[2], result__0_06, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[3], result__0_10, rtol=rtol, atol=atol)

    # Reminder: there's no call to do SCF here because we already have
    # the MO coefficients.
    mol = molecules.molecule_water_sto3g()
    mol.build()
    polarizability = electric.Polarizability(
        Program.PySCF,
        mol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
    )
    polarizability.form_operators()
    polarizability.run(hamiltonian=hamiltonian, spin=spin)
    polarizability.form_results()

    np.testing.assert_allclose(
        polarizability.polarizabilities[0], result__0_00, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[1], result__0_02, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[2], result__0_06, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[3], result__0_10, rtol=rtol, atol=atol
    )


def test_final_result_rhf_h2o_sto3g_rpa_triplet() -> None:
    """Test correctness of the final result for water/STO-3G with full RPA for
    triplet response induced by the dipole length operator computed with
    quantities from disk.
    """
    hamiltonian = Hamiltonian.RPA
    spin = Spin.triplet

    C = utils.fix_mocoeffs_shape(utils.np_load(REFDIR / "C.npz"))
    E = utils.fix_moenergies_shape(utils.np_load(REFDIR / "F_MO.npz"))
    TEI_MO = utils.np_load(REFDIR / "TEI_MO.npz")
    # nocc_alph, nvirt_alph, nocc_beta, nvirt_beta
    occupations = np.asarray([5, 2, 5, 2], dtype=int)
    stub = "h2o_sto3g_"
    dim = occupations[0] + occupations[1]
    mat_dipole_x = utils.parse_int_file_2(REFDIR / f"{stub}mux.dat", dim)
    mat_dipole_y = utils.parse_int_file_2(REFDIR / f"{stub}muy.dat", dim)
    mat_dipole_z = utils.parse_int_file_2(REFDIR / f"{stub}muz.dat", dim)

    solver = solvers.ExactInv(C, E, occupations)
    solver.tei_mo = (TEI_MO,)
    solver.tei_mo_type = AO2MOTransformationType.full
    driver = cphf.CPHF(solver)
    ao_integrals_dipole = np.stack((mat_dipole_x, mat_dipole_y, mat_dipole_z), axis=0)
    operator_dipole = operators.Operator(
        label="dipole", is_imaginary=False, is_spin_dependent=False
    )
    operator_dipole.ao_integrals = ao_integrals_dipole
    driver.add_operator(operator_dipole)
    frequencies = (0.0, 0.02, 0.06, 0.1)
    driver.set_frequencies(frequencies)

    driver.run(hamiltonian=hamiltonian, spin=spin, program=None, program_obj=None)

    assert len(driver.results) == len(frequencies)

    result__0_00 = np.array(
        [[26.59744305, 0.0, 0.0], [0.0, 18.11879557, 0.0], [0.0, 0.0, 0.07798969]]
    )
    result__0_02 = np.array(
        [[26.68282287, 0.0, 0.0], [0.0, 18.19390051, 0.0], [0.0, 0.0, 0.07837521]]
    )
    result__0_06 = np.array(
        [[27.38617401, 0.0, 0.0], [0.0, 18.81922578, 0.0], [0.0, 0.0, 0.08160226]]
    )
    result__0_10 = np.array(
        [[28.91067234, 0.0, 0.0], [0.0, 20.21670386, 0.0], [0.0, 0.0, 0.08892512]]
    )

    atol = 1.0e-8
    rtol = 0.0
    np.testing.assert_allclose(driver.results[0], result__0_00, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[1], result__0_02, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[2], result__0_06, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[3], result__0_10, rtol=rtol, atol=atol)

    mol = molecules.molecule_water_sto3g()
    mol.build()
    polarizability = electric.Polarizability(
        Program.PySCF,
        mol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
    )
    polarizability.form_operators()
    polarizability.run(hamiltonian=hamiltonian, spin=spin)
    polarizability.form_results()

    np.testing.assert_allclose(
        polarizability.polarizabilities[0], result__0_00, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[1], result__0_02, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[2], result__0_06, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[3], result__0_10, rtol=rtol, atol=atol
    )


def test_final_result_rhf_h2o_sto3g_tda_singlet() -> None:
    """Test correctness of the final result for water/STO-3G with the TDA
    approximation/CIS for singlet response induced by the dipole length
    operator computed with quantities from disk.
    """
    hamiltonian = Hamiltonian.TDA
    spin = Spin.singlet

    C = utils.fix_mocoeffs_shape(utils.np_load(REFDIR / "C.npz"))
    E = utils.fix_moenergies_shape(utils.np_load(REFDIR / "F_MO.npz"))
    TEI_MO = utils.np_load(REFDIR / "TEI_MO.npz")
    # nocc_alph, nvirt_alph, nocc_beta, nvirt_beta
    occupations = np.asarray([5, 2, 5, 2], dtype=int)
    stub = "h2o_sto3g_"
    dim = occupations[0] + occupations[1]
    mat_dipole_x = utils.parse_int_file_2(REFDIR / f"{stub}mux.dat", dim)
    mat_dipole_y = utils.parse_int_file_2(REFDIR / f"{stub}muy.dat", dim)
    mat_dipole_z = utils.parse_int_file_2(REFDIR / f"{stub}muz.dat", dim)

    solver = solvers.ExactInv(C, E, occupations)
    solver.tei_mo = (TEI_MO,)
    solver.tei_mo_type = AO2MOTransformationType.full
    driver = cphf.CPHF(solver)
    ao_integrals_dipole = np.stack((mat_dipole_x, mat_dipole_y, mat_dipole_z), axis=0)
    operator_dipole = operators.Operator(
        label="dipole", is_imaginary=False, is_spin_dependent=False
    )
    operator_dipole.ao_integrals = ao_integrals_dipole
    driver.add_operator(operator_dipole)
    frequencies = (0.0, 0.02, 0.06, 0.1)
    driver.set_frequencies(frequencies)

    driver.run(hamiltonian=hamiltonian, spin=spin, program=None, program_obj=None)

    assert len(driver.results) == len(frequencies)

    result__0_00 = np.array([[8.89855952, 0.0, 0.0], [0.0, 4.00026556, 0.0], [0.0, 0.0, 0.0552774]])
    result__0_02 = np.array(
        [[8.90690928, 0.0, 0.0], [0.0, 4.00298342, 0.0], [0.0, 0.0, 0.05545196]]
    )
    result__0_06 = np.array(
        [[8.97427725, 0.0, 0.0], [0.0, 4.02491517, 0.0], [0.0, 0.0, 0.05688918]]
    )
    result__0_10 = np.array(
        [[9.11212633, 0.0, 0.0], [0.0, 4.06981937, 0.0], [0.0, 0.0, 0.05999934]]
    )

    atol = 1.0e-8
    rtol = 0.0
    np.testing.assert_allclose(driver.results[0], result__0_00, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[1], result__0_02, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[2], result__0_06, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[3], result__0_10, rtol=rtol, atol=atol)

    mol = molecules.molecule_water_sto3g()
    mol.build()
    polarizability = electric.Polarizability(
        Program.PySCF,
        mol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
    )
    polarizability.form_operators()
    polarizability.run(hamiltonian=hamiltonian, spin=spin)
    polarizability.form_results()

    np.testing.assert_allclose(
        polarizability.polarizabilities[0], result__0_00, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[1], result__0_02, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[2], result__0_06, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[3], result__0_10, rtol=rtol, atol=atol
    )


def test_final_result_rhf_h2o_sto3g_tda_triplet() -> None:
    """Test correctness of the final result for water/STO-3G with the TDA
    approximation/CIS for triplet response induced by the dipole length
    operator computed with quantities from disk.
    """
    hamiltonian = Hamiltonian.TDA
    spin = Spin.triplet

    C = utils.fix_mocoeffs_shape(utils.np_load(REFDIR / "C.npz"))
    E = utils.fix_moenergies_shape(utils.np_load(REFDIR / "F_MO.npz"))
    TEI_MO = utils.np_load(REFDIR / "TEI_MO.npz")
    # nocc_alph, nvirt_alph, nocc_beta, nvirt_beta
    occupations = np.asarray([5, 2, 5, 2], dtype=int)
    stub = "h2o_sto3g_"
    dim = occupations[0] + occupations[1]
    mat_dipole_x = utils.parse_int_file_2(REFDIR / f"{stub}mux.dat", dim)
    mat_dipole_y = utils.parse_int_file_2(REFDIR / f"{stub}muy.dat", dim)
    mat_dipole_z = utils.parse_int_file_2(REFDIR / f"{stub}muz.dat", dim)

    solver = solvers.ExactInv(C, E, occupations)
    solver.tei_mo = (TEI_MO,)
    solver.tei_mo_type = AO2MOTransformationType.full
    driver = cphf.CPHF(solver)
    ao_integrals_dipole = np.stack((mat_dipole_x, mat_dipole_y, mat_dipole_z), axis=0)
    operator_dipole = operators.Operator(
        label="dipole", is_imaginary=False, is_spin_dependent=False
    )
    operator_dipole.ao_integrals = ao_integrals_dipole
    driver.add_operator(operator_dipole)
    frequencies = (0.0, 0.02, 0.06, 0.1)
    driver.set_frequencies(frequencies)

    driver.run(hamiltonian=hamiltonian, spin=spin, program=None, program_obj=None)

    assert len(driver.results) == len(frequencies)

    result__0_00 = np.array(
        [[14.64430714, 0.0, 0.0], [0.0, 8.80921432, 0.0], [0.0, 0.0, 0.06859496]]
    )
    result__0_02 = np.array(
        [[14.68168443, 0.0, 0.0], [0.0, 8.83562647, 0.0], [0.0, 0.0, 0.0689291]]
    )
    result__0_06 = np.array(
        [[14.98774296, 0.0, 0.0], [0.0, 9.0532224, 0.0], [0.0, 0.0, 0.07172414]]
    )
    result__0_10 = np.array(
        [[15.63997724, 0.0, 0.0], [0.0, 9.52504267, 0.0], [0.0, 0.0, 0.07805428]]
    )

    atol = 1.0e-8
    rtol = 0.0
    np.testing.assert_allclose(driver.results[0], result__0_00, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[1], result__0_02, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[2], result__0_06, rtol=rtol, atol=atol)
    np.testing.assert_allclose(driver.results[3], result__0_10, rtol=rtol, atol=atol)

    mol = molecules.molecule_water_sto3g()
    mol.build()
    polarizability = electric.Polarizability(
        Program.PySCF,
        mol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
    )
    polarizability.form_operators()
    polarizability.run(hamiltonian=hamiltonian, spin=spin)
    polarizability.form_results()

    np.testing.assert_allclose(
        polarizability.polarizabilities[0], result__0_00, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[1], result__0_02, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[2], result__0_06, rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        polarizability.polarizabilities[3], result__0_10, rtol=rtol, atol=atol
    )


def test_as_many_as_possible_rhf_disk() -> None:
    """Test correctness of the final result for closed-shell molecules against
    DALTON references.
    """
    run_as_many_tests_as_possible_rhf_disk("r_lih_hf_sto-3g")


def test_as_many_as_possible_uhf_disk() -> None:
    """Test correctness of the final result for open-shell (UHF) molecules
    against DALTON (ROHF) references.
    """
    run_as_many_tests_as_possible_uhf_disk("u_lih_cation_hf_sto-3g")



from abc import ABC, abstractmethod
from typing import Callable, Optional, Sequence, Tuple

import numpy as np
import scipy as sp

from pymolresponse.core import AO2MOTransformationType, Hamiltonian, Program, Spin
from pymolresponse.explicit_equations_full import (
    form_rpa_a_matrix_mo_singlet_full,
    form_rpa_a_matrix_mo_singlet_os_full,
    form_rpa_a_matrix_mo_singlet_ss_full,
    form_rpa_a_matrix_mo_triplet_full,
    form_rpa_b_matrix_mo_singlet_full,
    form_rpa_b_matrix_mo_singlet_os_full,
    form_rpa_b_matrix_mo_singlet_ss_full,
    form_rpa_b_matrix_mo_triplet_full,
)
from pymolresponse.explicit_equations_partial import (
    form_rpa_a_matrix_mo_singlet_os_partial,
    form_rpa_a_matrix_mo_singlet_partial,
    form_rpa_a_matrix_mo_singlet_ss_partial,
    form_rpa_a_matrix_mo_triplet_partial,
    form_rpa_b_matrix_mo_singlet_os_partial,
    form_rpa_b_matrix_mo_singlet_partial,
    form_rpa_b_matrix_mo_singlet_ss_partial,
    form_rpa_b_matrix_mo_triplet_partial,
)
from pymolresponse.integrals import JK
from pymolresponse.operators import Operator
from pymolresponse.utils import repack_matrix_to_vector

np.set_printoptions(precision=5, linewidth=200, suppress=True)


class Solver(ABC):
    """A Solver does all of the TODO

    There are two principal algorithm classes for forming response vectors:

    - Exact, where the full electronic Hessian is formed explicitly,
      diagonalized, and then contracted with property gradient vectors to
      response vectors, or

    - Iterative, where the electronic Hessian is never formed explicitly, TODO
    """

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        # MO coefficients: the first axis is alpha/beta
        assert len(mocoeffs.shape) == 3
        assert (mocoeffs.shape[0] == 1) or (mocoeffs.shape[0] == 2)
        self.is_uhf = mocoeffs.shape[0] == 2
        # MO energies: the first axis is alpha/beta, the other two represent a
        # diagonal square matrix.
        assert len(moenergies.shape) == 3
        assert (moenergies.shape[0] == 1) or (moenergies.shape[0] == 2)
        if self.is_uhf:
            assert moenergies.shape[0] == 2
        else:
            assert moenergies.shape[0] == 1
        assert moenergies.shape[1] == moenergies.shape[2]

        assert len(occupations) == 4

        self.mocoeffs = mocoeffs
        self.moenergies = moenergies
        self.occupations = occupations
        self.form_ranges_from_occupations()

        self.operators = []
        self.frequencies = []

        self.indices_closed_act = None
        self.indices_closed_secondary = None
        self.indices_act_secondary = None
        self.indices_rohf = None
        self.indices_display_rohf = None

        # These are needed for MO-based solvers.
        self.tei_mo = None
        self.tei_mo_type = AO2MOTransformationType.partial
        self.explicit_hessian = None
        self.explicit_hessian_inv = None

    def form_tei_mo(
        self, program: Program, program_obj, tei_mo_type: AO2MOTransformationType
    ) -> None:
        assert isinstance(program, Program)
        # TODO program_obj
        assert isinstance(tei_mo_type, AO2MOTransformationType)
        nden = self.mocoeffs.shape[0]
        assert nden in (1, 2)
        if program == Program.PySCF:
            from pymolresponse.interfaces.pyscf.ao2mo import AO2MOpyscf

            ao2mo = AO2MOpyscf(self.mocoeffs, program_obj.verbose, program_obj)
        elif program == Program.Psi4:
            import psi4

            from pymolresponse.ao2mo import AO2MO

            assert isinstance(program_obj, psi4.core.Molecule)
            ao2mo = AO2MO(
                self.mocoeffs,
                self.occupations,
                I=psi4.core.MintsHelper(
                    psi4.core.Wavefunction.build(
                        program_obj, psi4.core.get_global_option("BASIS")
                    ).basisset()
                )
                .ao_eri()
                .np,
            )
        else:
            raise RuntimeError
        if tei_mo_type == AO2MOTransformationType.partial and nden == 2:
            ao2mo.perform_uhf_partial()
        elif tei_mo_type == AO2MOTransformationType.partial and nden == 1:
            ao2mo.perform_rhf_partial()
        elif tei_mo_type == AO2MOTransformationType.full and nden == 2:
            ao2mo.perform_uhf_full()
        elif tei_mo_type == AO2MOTransformationType.full and nden == 1:
            ao2mo.perform_rhf_full()
        else:
            raise ValueError(f"Unknown combination of tei_mo_type {tei_mo_type} and nden {nden}")
        self.tei_mo = ao2mo.tei_mo
        self.tei_mo_type = tei_mo_type

    def form_ranges_from_occupations(self) -> None:
        assert len(self.occupations) == 4
        nocc_a, nvirt_a, nocc_b, nvirt_b = self.occupations
        assert (nocc_a + nvirt_a) == (nocc_b + nvirt_b)
        norb = nocc_a + nvirt_a
        nelec = nocc_a + nocc_b
        nact = abs(int(nocc_a - nocc_b))
        nclosed = (nelec - nact) // 2
        nsecondary = norb - (nclosed + nact)
        range_closed = list(range(0, nclosed))
        range_act = list(range(nclosed, nclosed + nact))
        range_secondary = list(range(nclosed + nact, nclosed + nact + nsecondary))
        self.indices_closed_act = [(i, t) for i in range_closed for t in range_act]
        self.indices_closed_secondary = [(i, a) for i in range_closed for a in range_secondary]
        self.indices_act_secondary = [(t, a) for t in range_act for a in range_secondary]
        self.indices_rohf = (
            self.indices_closed_act + self.indices_closed_secondary + self.indices_act_secondary
        )
        self.indices_display_rohf = [(p + 1, q + 1) for (p, q) in self.indices_rohf]

    def set_frequencies(self, frequencies: Optional[Sequence[float]] = None) -> None:
        if frequencies is None:
            self.frequencies = [0.0]
        else:
            self.frequencies = frequencies
        if hasattr(self, "operators"):
            for operator in self.operators:
                operator.frequencies = self.frequencies

    def add_operator(self, operator: Operator) -> None:
        # First dimension is the number of Cartesian components, next
        # two are the number of AOs.
        assert hasattr(operator, "ao_integrals")
        shape = operator.ao_integrals.shape
        assert len(shape) == 3
        assert shape[0] >= 1
        assert shape[1] == shape[2]
        operator.indices_closed_act = self.indices_closed_act
        operator.indices_closed_secondary = self.indices_closed_secondary
        operator.indices_act_secondary = self.indices_act_secondary
        # Form the property gradient.
        operator.form_rhs(self.mocoeffs, self.occupations)
        self.operators.append(operator)

    @abstractmethod
    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        """Run the solver."""


class LineqSolver(Solver, ABC):
    """Base class for all solvers of the type $AX = B$."""

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)


class ExactLineqSolver(LineqSolver, ABC):
    """Base class for all solvers of the type $AX = B$, where $A$ is formed
    explicitly.
    """

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> np.ndarray:
        super().__init__(mocoeffs, moenergies, occupations)

    def form_explicit_hessian(self, hamiltonian: Hamiltonian, spin: Spin, frequency: float) -> None:
        assert self.tei_mo is not None
        assert len(self.tei_mo) in (1, 2, 4, 6)
        assert isinstance(self.tei_mo_type, AO2MOTransformationType)

        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(frequency, (float, type(None)))

        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.occupations
        nov_alph = nocc_alph * nvirt_alph
        nov_beta = nocc_beta * nvirt_beta

        superoverlap_alph = (
            np.block(
                [
                    [np.eye(nov_alph), np.zeros(shape=(nov_alph, nov_alph))],
                    [np.zeros(shape=(nov_alph, nov_alph)), -np.eye(nov_alph)],
                ]
            )
            * frequency
        )

        if not self.is_uhf:
            # Set up "function pointers".
            if self.tei_mo_type == AO2MOTransformationType.full:
                assert len(self.tei_mo) == 1
                tei_mo = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                assert len(self.tei_mo) == 2
                tei_mo_ovov = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
                tei_mo_oovv = self.tei_mo[1]  # ty: ignore[index-out-of-bounds]

            if self.tei_mo_type == AO2MOTransformationType.full:
                if hamiltonian == Hamiltonian.RPA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = form_rpa_b_matrix_mo_singlet_full(tei_mo, nocc_alph)
                elif hamiltonian == Hamiltonian.RPA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = form_rpa_b_matrix_mo_triplet_full(tei_mo, nocc_alph)
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = np.zeros(shape=(nov_alph, nov_alph))
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = np.zeros(shape=(nov_alph, nov_alph))
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                if hamiltonian == Hamiltonian.RPA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_partial(
                        self.moenergies[0], tei_mo_ovov, tei_mo_oovv
                    )
                    B = form_rpa_b_matrix_mo_singlet_partial(tei_mo_ovov)
                elif hamiltonian == Hamiltonian.RPA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_partial(self.moenergies[0], tei_mo_oovv)
                    B = form_rpa_b_matrix_mo_triplet_partial(tei_mo_ovov)
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_partial(
                        self.moenergies[0], tei_mo_ovov, tei_mo_oovv
                    )
                    B = np.zeros(shape=(nov_alph, nov_alph))
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_partial(self.moenergies[0], tei_mo_oovv)
                    B = np.zeros(shape=(nov_alph, nov_alph))

            G = np.block([[A, B], [B, A]])
            self.explicit_hessian = G - superoverlap_alph

        else:
            # For UHF there are both "operator-dependent" and
            # operator-indepenent parts of the orbital Hessian because
            # the opposite-spin property gradient couples in. Here,
            # only form the 4 blocks of the "super-Hessian" (a
            # supermatrix of supermatrices); the equations will get
            # pieced together when it is time ot form the response
            # vectors.

            # Set up "function pointers".
            if self.tei_mo_type == AO2MOTransformationType.full:
                assert len(self.tei_mo) == 4
                tei_mo_aaaa = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
                tei_mo_aabb = self.tei_mo[1]  # ty: ignore[index-out-of-bounds]
                tei_mo_bbaa = self.tei_mo[2]  # ty: ignore[index-out-of-bounds]
                tei_mo_bbbb = self.tei_mo[3]  # ty: ignore[index-out-of-bounds]
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                assert len(self.tei_mo) == 6
                tei_mo_ovov_aaaa = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
                tei_mo_ovov_aabb = self.tei_mo[1]  # ty: ignore[index-out-of-bounds]
                tei_mo_ovov_bbaa = self.tei_mo[2]  # ty: ignore[index-out-of-bounds]
                tei_mo_ovov_bbbb = self.tei_mo[3]  # ty: ignore[index-out-of-bounds]
                tei_mo_oovv_aaaa = self.tei_mo[4]  # ty: ignore[index-out-of-bounds]
                tei_mo_oovv_bbbb = self.tei_mo[5]  # ty: ignore[index-out-of-bounds]
            else:
                pass

            E_a = self.moenergies[0]
            E_b = self.moenergies[1]

            # TODO clean this up...
            if self.tei_mo_type == AO2MOTransformationType.full:
                if hamiltonian == Hamiltonian.RPA and spin == Spin.singlet:
                    A_ss_a = form_rpa_a_matrix_mo_singlet_ss_full(E_a, tei_mo_aaaa, nocc_alph)
                    A_os_a = form_rpa_a_matrix_mo_singlet_os_full(tei_mo_aabb, nocc_alph, nocc_beta)
                    B_ss_a = form_rpa_b_matrix_mo_singlet_ss_full(tei_mo_aaaa, nocc_alph)
                    B_os_a = form_rpa_b_matrix_mo_singlet_os_full(tei_mo_aabb, nocc_alph, nocc_beta)
                    A_ss_b = form_rpa_a_matrix_mo_singlet_ss_full(E_b, tei_mo_bbbb, nocc_beta)
                    A_os_b = form_rpa_a_matrix_mo_singlet_os_full(tei_mo_bbaa, nocc_beta, nocc_alph)
                    B_ss_b = form_rpa_b_matrix_mo_singlet_ss_full(tei_mo_bbbb, nocc_beta)
                    B_os_b = form_rpa_b_matrix_mo_singlet_os_full(tei_mo_bbaa, nocc_beta, nocc_alph)
                elif hamiltonian == Hamiltonian.RPA and spin == Spin.triplet:
                    # Since the "triplet" part contains no Coulomb contribution, and
                    # (xx|yy) is only in the Coulomb part, there is no ss/os
                    # separation for the triplet part.
                    zeros_ab = np.zeros(shape=(nov_alph, nov_beta))
                    zeros_ba = zeros_ab.T
                    A_ss_a = form_rpa_a_matrix_mo_triplet_full(E_a, tei_mo_aaaa, nocc_alph)
                    A_os_a = zeros_ab
                    B_ss_a = form_rpa_b_matrix_mo_triplet_full(tei_mo_aaaa, nocc_alph)
                    B_os_a = zeros_ab
                    A_ss_b = form_rpa_a_matrix_mo_triplet_full(E_b, tei_mo_bbbb, nocc_beta)
                    A_os_b = zeros_ba
                    B_ss_b = form_rpa_b_matrix_mo_triplet_full(tei_mo_bbbb, nocc_beta)
                    B_os_b = zeros_ba
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.singlet:
                    zeros_aa = np.zeros(shape=(nov_alph, nov_alph))
                    zeros_ab = np.zeros(shape=(nov_alph, nov_beta))
                    zeros_ba = zeros_ab.T
                    zeros_bb = np.zeros(shape=(nov_beta, nov_beta))
                    A_ss_a = form_rpa_a_matrix_mo_singlet_ss_full(E_a, tei_mo_aaaa, nocc_alph)
                    A_os_a = form_rpa_a_matrix_mo_singlet_os_full(tei_mo_aabb, nocc_alph, nocc_beta)
                    B_ss_a = zeros_aa
                    B_os_a = zeros_ab
                    A_ss_b = form_rpa_a_matrix_mo_singlet_ss_full(E_b, tei_mo_bbbb, nocc_beta)
                    A_os_b = form_rpa_a_matrix_mo_singlet_os_full(tei_mo_bbaa, nocc_beta, nocc_alph)
                    B_ss_b = zeros_bb
                    B_os_b = zeros_ba
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.triplet:
                    zeros_aa = np.zeros(shape=(nov_alph, nov_alph))
                    zeros_ab = np.zeros(shape=(nov_alph, nov_beta))
                    zeros_ba = zeros_ab.T
                    zeros_bb = np.zeros(shape=(nov_beta, nov_beta))
                    A_ss_a = form_rpa_a_matrix_mo_triplet_full(E_a, tei_mo_aaaa, nocc_alph)
                    A_os_a = zeros_ab
                    B_ss_a = zeros_aa
                    B_os_a = zeros_ab
                    A_ss_b = form_rpa_a_matrix_mo_triplet_full(E_b, tei_mo_bbbb, nocc_beta)
                    A_os_b = zeros_ba
                    B_ss_b = zeros_bb
                    B_os_b = zeros_ba

            elif self.tei_mo_type == AO2MOTransformationType.partial:
                if hamiltonian == Hamiltonian.RPA and spin == Spin.singlet:
                    A_ss_a = form_rpa_a_matrix_mo_singlet_ss_partial(
                        E_a, tei_mo_ovov_aaaa, tei_mo_oovv_aaaa
                    )
                    A_os_a = form_rpa_a_matrix_mo_singlet_os_partial(tei_mo_ovov_aabb)
                    B_ss_a = form_rpa_b_matrix_mo_singlet_ss_partial(tei_mo_ovov_aaaa)
                    B_os_a = form_rpa_b_matrix_mo_singlet_os_partial(tei_mo_ovov_aabb)
                    A_ss_b = form_rpa_a_matrix_mo_singlet_ss_partial(
                        E_b, tei_mo_ovov_bbbb, tei_mo_oovv_bbbb
                    )
                    A_os_b = form_rpa_a_matrix_mo_singlet_os_partial(tei_mo_ovov_bbaa)
                    B_ss_b = form_rpa_b_matrix_mo_singlet_ss_partial(tei_mo_ovov_bbbb)
                    B_os_b = form_rpa_b_matrix_mo_singlet_os_partial(tei_mo_ovov_bbaa)
                elif hamiltonian == Hamiltonian.RPA and spin == Spin.triplet:
                    # Since the "triplet" part contains no Coulomb contribution, and
                    # (xx|yy) is only in the Coulomb part, there is no ss/os
                    # separation for the triplet part.
                    zeros_ab = np.zeros(shape=(nov_alph, nov_beta))
                    zeros_ba = zeros_ab.T
                    A_ss_a = form_rpa_a_matrix_mo_triplet_partial(E_a, tei_mo_oovv_aaaa)
                    A_os_a = zeros_ab
                    B_ss_a = form_rpa_b_matrix_mo_triplet_partial(tei_mo_ovov_aaaa)
                    B_os_a = zeros_ab
                    A_ss_b = form_rpa_a_matrix_mo_triplet_partial(E_b, tei_mo_oovv_bbbb)
                    A_os_b = zeros_ba
                    B_ss_b = form_rpa_b_matrix_mo_triplet_partial(tei_mo_ovov_bbbb)
                    B_os_b = zeros_ba
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.singlet:
                    zeros_aa = np.zeros(shape=(nov_alph, nov_alph))
                    zeros_ab = np.zeros(shape=(nov_alph, nov_beta))
                    zeros_ba = zeros_ab.T
                    zeros_bb = np.zeros(shape=(nov_beta, nov_beta))
                    A_ss_a = form_rpa_a_matrix_mo_singlet_ss_partial(
                        E_a, tei_mo_ovov_aaaa, tei_mo_oovv_aaaa
                    )
                    A_os_a = form_rpa_a_matrix_mo_singlet_os_partial(tei_mo_ovov_aabb)
                    B_ss_a = zeros_aa
                    B_os_a = zeros_ab
                    A_ss_b = form_rpa_a_matrix_mo_singlet_ss_partial(
                        E_b, tei_mo_ovov_bbbb, tei_mo_oovv_bbbb
                    )
                    A_os_b = form_rpa_a_matrix_mo_singlet_os_partial(tei_mo_ovov_bbaa)
                    B_ss_b = zeros_bb
                    B_os_b = zeros_ba
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.triplet:
                    zeros_aa = np.zeros(shape=(nov_alph, nov_alph))
                    zeros_ab = np.zeros(shape=(nov_alph, nov_beta))
                    zeros_ba = zeros_ab.T
                    zeros_bb = np.zeros(shape=(nov_beta, nov_beta))
                    A_ss_a = form_rpa_a_matrix_mo_triplet_partial(E_a, tei_mo_oovv_aaaa)
                    A_os_a = zeros_ab
                    B_ss_a = zeros_aa
                    B_os_a = zeros_ab
                    A_ss_b = form_rpa_a_matrix_mo_triplet_partial(E_b, tei_mo_oovv_bbbb)
                    A_os_b = zeros_ba
                    B_ss_b = zeros_bb
                    B_os_b = zeros_ba

            superoverlap_beta = np.block(
                [
                    [np.eye(nov_beta), np.zeros(shape=(nov_beta, nov_beta))],
                    [np.zeros(shape=(nov_beta, nov_beta)), -np.eye(nov_beta)],
                ]
            )
            superoverlap_beta = superoverlap_beta * frequency

            G_aa = np.block([[A_ss_a, B_ss_a], [B_ss_a, A_ss_a]])
            G_ab = np.block([[A_os_a, B_os_a], [B_os_a, A_os_a]])
            G_ba = np.block([[A_os_b, B_os_b], [B_os_b, A_os_b]])
            G_bb = np.block([[A_ss_b, B_ss_b], [B_ss_b, A_ss_b]])

            self.explicit_hessian = (G_aa - superoverlap_alph, G_ab, G_ba, G_bb - superoverlap_beta)

    @abstractmethod
    def invert_explicit_hessian(self) -> None:
        """Invert the full explicitly-constructed electronic Hessian."""

    def form_response_vectors(self) -> None:
        if self.is_uhf:
            assert isinstance(self.explicit_hessian, tuple)
            assert len(self.explicit_hessian) == 4
            # TODO
            # assert isinstance(self.explicit_hessian_inv, tuple)
            # assert len(self.explicit_hessian_inv) == 2
            G_aa, G_ab, G_ba, G_bb = self.explicit_hessian
            G_aa_inv, G_bb_inv = self.explicit_hessian_inv  # ty: ignore[not-iterable]
        for operator in self.operators:
            if not self.is_uhf:
                rspvecs_operator = []
                for idx_operator_component in range(operator.ao_integrals.shape[0]):
                    shape = operator.mo_integrals_ai_supervector_alph[idx_operator_component].shape
                    assert len(shape) == 2
                    assert shape[1] == 1
                    rspvec_operator_component = np.dot(
                        self.explicit_hessian_inv,
                        operator.mo_integrals_ai_supervector_alph[idx_operator_component],
                    )
                    assert rspvec_operator_component.shape == shape
                    rspvecs_operator.append(rspvec_operator_component)
                # TODO this isn't working and I don't know why
                # rspvecs_operator = np.stack(rspvecs_operator, axis=0)

                # All the lines with 'tmp' could be replaced by a working
                # stack call.
                tmp = np.empty(
                    shape=(len(rspvecs_operator), shape[0], 1),
                    dtype=rspvec_operator_component.dtype,
                )
                for idx, rspvec_operator_component in enumerate(rspvecs_operator):
                    tmp[idx] = rspvec_operator_component
                rspvecs_operator = tmp
                operator.rspvecs_alph.append(rspvecs_operator)
            else:
                # Form the operator-dependent part of the response vectors.
                rspvecs_operator_alph = []
                rspvecs_operator_beta = []
                for idx_operator_component in range(operator.ao_integrals.shape[0]):
                    operator_component_alph = operator.mo_integrals_ai_supervector_alph[
                        idx_operator_component
                    ]
                    operator_component_beta = operator.mo_integrals_ai_supervector_beta[
                        idx_operator_component
                    ]
                    shape_alph = operator_component_alph.shape
                    shape_beta = operator_component_beta.shape
                    assert len(shape_alph) == len(shape_beta) == 2
                    assert shape_alph[1] == shape_beta[1] == 1
                    right_alph = operator_component_alph - (
                        np.dot(G_ab, np.dot(G_bb_inv, operator_component_beta))
                    )
                    right_beta = operator_component_beta - (
                        np.dot(G_ba, np.dot(G_aa_inv, operator_component_alph))
                    )
                    assert right_alph.shape == shape_alph
                    assert right_beta.shape == shape_beta
                    rspvecs_operator_alph.append(np.dot(self.left_alph, right_alph))
                    rspvecs_operator_beta.append(np.dot(self.left_beta, right_beta))
                tmp_alph = np.empty(
                    shape=(len(rspvecs_operator_alph), shape_alph[0], 1),
                    dtype=operator_component_alph.dtype,
                )
                tmp_beta = np.empty(
                    shape=(len(rspvecs_operator_beta), shape_beta[0], 1),
                    dtype=operator_component_beta.dtype,
                )
                for idx_alph, operator_component_alph in enumerate(rspvecs_operator_alph):
                    tmp_alph[idx_alph] = operator_component_alph
                for idx_beta, operator_component_beta in enumerate(rspvecs_operator_beta):
                    tmp_beta[idx_beta] = operator_component_beta
                rspvecs_operator_alph = tmp_alph
                rspvecs_operator_beta = tmp_beta
                operator.rspvecs_alph.append(rspvecs_operator_alph)
                operator.rspvecs_beta.append(rspvecs_operator_beta)

    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(program, (Program, type(None)))
        # TODO program_obj
        if not self.tei_mo:
            self.form_tei_mo(program, program_obj, AO2MOTransformationType.partial)
        for frequency in self.frequencies:
            self.form_explicit_hessian(hamiltonian, spin, frequency)
            self.invert_explicit_hessian()
            self.form_response_vectors()


class ExactInv(ExactLineqSolver):
    def __init__(
        self,
        mocoeffs: np.ndarray,
        moenergies: np.ndarray,
        occupations: np.ndarray,
        *,
        inv_func: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)

        if inv_func is not None:
            self.inv_func = inv_func
        else:
            self.inv_func = np.linalg.inv

    def invert_explicit_hessian(self) -> None:
        assert hasattr(self, "explicit_hessian")
        if not self.is_uhf:
            assert isinstance(self.explicit_hessian, np.ndarray)
            assert len(self.explicit_hessian.shape) == 2  # ty: ignore[invalid-argument-type]
            assert self.explicit_hessian.shape[0] == self.explicit_hessian.shape[1]  # ty: ignore[not-subscriptable]
            self.explicit_hessian_inv = self.inv_func(self.explicit_hessian)
        else:
            assert isinstance(self.explicit_hessian, tuple)
            assert len(self.explicit_hessian) == 4
            self.explicit_hessian_inv = []
            G_aa, G_ab, G_ba, G_bb = self.explicit_hessian
            # The inverse of the opposite-spin blocks is not
            # necessary.
            G_aa_inv = self.inv_func(G_aa)
            G_bb_inv = self.inv_func(G_bb)
            self.explicit_hessian_inv.append(G_aa_inv)
            self.explicit_hessian_inv.append(G_bb_inv)
            # Form the operator-independent part of the response
            # vectors.
            self.left_alph = self.inv_func(G_aa - np.dot(G_ab, np.dot(G_bb_inv, G_ba)))
            self.left_beta = self.inv_func(G_bb - np.dot(G_ba, np.dot(G_aa_inv, G_ab)))


class ExactInvCholesky(ExactLineqSolver):
    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)

    def invert_explicit_hessian(self) -> None:
        assert hasattr(self, "explicit_hessian")
        if not self.is_uhf:
            fac = sp.linalg.cholesky(self.explicit_hessian)
            fac_inv = sp.linalg.inv(fac)
            inv = np.dot(fac_inv, fac_inv.T)
            self.explicit_hessian_inv = inv
        else:
            assert isinstance(self.explicit_hessian, tuple)
            assert len(self.explicit_hessian) == 4
            self.explicit_hessian_inv = []
            G_aa, G_ab, G_ba, G_bb = self.explicit_hessian
            # The inverse of the opposite-spin blocks is not
            # necessary.
            G_aa_R_inv = sp.linalg.inv(sp.linalg.cholesky(G_aa, lower=False))
            G_bb_R_inv = sp.linalg.inv(sp.linalg.cholesky(G_bb, lower=False))
            G_aa_inv = np.dot(G_aa_R_inv, G_aa_R_inv.T)
            G_bb_inv = np.dot(G_bb_R_inv, G_bb_R_inv.T)
            self.explicit_hessian_inv.append(G_aa_inv)
            self.explicit_hessian_inv.append(G_bb_inv)
            # Form the operator-independent part of the response
            # vectors.
            left_alph_R = sp.linalg.cholesky(
                G_aa - np.dot(G_ab, np.dot(G_bb_inv, G_ba)), lower=False
            )
            left_beta_R = sp.linalg.cholesky(
                G_bb - np.dot(G_ba, np.dot(G_aa_inv, G_ab)), lower=False
            )
            left_alph_R_inv = sp.linalg.inv(left_alph_R)
            left_beta_R_inv = sp.linalg.inv(left_beta_R)
            self.left_alph = np.dot(left_alph_R_inv, left_alph_R_inv.T)
            self.left_beta = np.dot(left_beta_R_inv, left_beta_R_inv.T)


class IterativeLinEqSolver(LineqSolver):
    def __init__(
        self,
        mocoeffs: np.ndarray,
        moenergies: np.ndarray,
        occupations: np.ndarray,
        jk_generator: JK,
        *,
        maxiter: int = 40,
        conv: float = 1.0e-9,
    ) -> None:
        assert isinstance(jk_generator, JK)
        assert isinstance(maxiter, int)
        assert isinstance(conv, float)
        super().__init__(mocoeffs, moenergies, occupations)

        # TODO
        self.jk_generator = jk_generator
        self.maxiter = maxiter
        self.conv = conv

    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(program, (Program, type(None)))
        # TODO program_obj
        assert self.frequencies
        for frequency in self.frequencies:
            self._run(hamiltonian, spin, frequency, self.maxiter, self.conv)

    def _run(
        self, hamiltonian: Hamiltonian, spin: Spin, omega: float, maxiter: int, conv: float
    ) -> None:
        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.occupations
        assert (nocc_alph + nvirt_alph) == (nocc_beta + nvirt_beta)
        norb = nocc_alph + nvirt_alph
        if self.is_uhf:
            raise RuntimeError
        else:
            epsilon = np.diag(self.moenergies[0])
            ia_denom = epsilon[nocc_alph:] - epsilon[:nocc_alph].reshape(-1, 1) - omega
            all_denom = epsilon.reshape(-1, 1) - epsilon - omega
            C = self.mocoeffs[0]
            Co = C[:, :nocc_alph]
            for operator in self.operators:
                ncomponents = operator.ao_integrals.shape[0]
                x = []
                x_old = []
                # TODO implement DIIS
                rhsmats = [
                    C.T @ operator.ao_integrals[component] @ C for component in range(ncomponents)
                ]
                for rhsmat in rhsmats:
                    U = np.zeros((norb, norb))
                    U[:nocc_alph, nocc_alph:] = rhsmat[:nocc_alph, nocc_alph:] / ia_denom
                    U[nocc_alph:, :nocc_alph] = rhsmat[nocc_alph:, :nocc_alph] / -ia_denom.T
                    x.append(U)
                    x_old.append(np.zeros_like(U))
                for i in range(1, maxiter + 1):
                    for component in range(ncomponents):
                        C_left_1 = Co
                        C_right_1 = C.dot(x[component].T)[:, :nocc_alph]
                        C_left_2 = (-C).dot(x[component])[:, :nocc_alph]
                        C_right_2 = Co
                        J_1, K_1 = self.jk_generator.compute_from_mocoeffs(C_left_1, C_right_1)
                        J_2, K_2 = self.jk_generator.compute_from_mocoeffs(C_left_2, C_right_2)
                        U = x[component].copy()
                        upd = (
                            rhsmats[component]
                            - (
                                x[component] * all_denom
                                - (C.T).dot(2 * J_1 - K_1 + 2 * J_2 - K_2).dot(C)
                            )
                        ) / all_denom
                        U[:nocc_alph, nocc_alph:] += upd[:nocc_alph, nocc_alph:]
                        U[nocc_alph:, :nocc_alph] += upd[nocc_alph:, :nocc_alph]
                        x[component] = U.copy()
                    rms = []
                    for component in range(ncomponents):
                        rms.append(np.max(x[component] - x_old[component]) ** 2)
                        x_old[component] = x[component]
                    avg_rms = sum(rms) / len(rms)
                    max_rms = max(rms)
                    if max_rms < conv:
                        print(f"CPHF converged in {i:d} iterations.")
                        operator.rspvecs_alph.append(
                            -np.stack(
                                [
                                    np.concatenate(
                                        (
                                            repack_matrix_to_vector(
                                                x[component][:nocc_alph, nocc_alph:].T
                                            ),
                                            repack_matrix_to_vector(
                                                x[component][nocc_alph:, :nocc_alph]
                                            ),
                                        )
                                    )
                                    for component in range(ncomponents)
                                ],
                                axis=0,
                            )[..., np.newaxis]
                        )
                        break
                    print(
                        f"CPHF Iteration {i:3d}: Average RMS = {avg_rms:3.8f}  Maximum RMS = {max_rms:3.8f}"
                    )


class EigSolver(Solver, ABC):
    """Base class for all solvers of the type $AX = 0$ (eigensolvers)."""

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)

    @staticmethod
    def norm_xy(z: np.ndarray, nocc: int, nvirt: int) -> Tuple[np.ndarray, np.ndarray]:
        x, y = z.reshape(2, nvirt, nocc)
        norm = 2 * (np.linalg.norm(x) ** 2 - np.linalg.norm(y) ** 2)
        norm = 1 / np.sqrt(norm)
        return (x * norm).flatten(), (y * norm).flatten()


class EigSolverTDA(EigSolver, ABC):
    """Base class for eigensolvers that use the Tamm-Dancoff approximation
    (TDA) rather than the random phase approximation (RPA).
    """

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)


class ExactDiagonalizationSolver(EigSolver):
    """Get the eigenvectors and eigenvalues of the RPA Hamiltonian by forming the
    electronic Hessian explicitly and diagonalizing it exactly.
    """

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)

    def form_explicit_hessian(self, hamiltonian: Hamiltonian, spin: Spin, frequency: float) -> None:
        assert hasattr(self, "tei_mo")
        assert self.tei_mo is not None
        assert len(self.tei_mo) in (1, 2, 4, 6)
        assert isinstance(self.tei_mo_type, AO2MOTransformationType)

        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(frequency, (float, type(None)))

        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.occupations
        nov_alph = nocc_alph * nvirt_alph
        nov_beta = nocc_beta * nvirt_beta  # noqa: F841

        if not self.is_uhf:
            # Set up "function pointers".
            if self.tei_mo_type == AO2MOTransformationType.full:
                assert len(self.tei_mo) == 1
                tei_mo = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                assert len(self.tei_mo) == 2
                tei_mo_ovov = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
                tei_mo_oovv = self.tei_mo[1]  # ty: ignore[index-out-of-bounds]

            if self.tei_mo_type == AO2MOTransformationType.full:
                if hamiltonian == Hamiltonian.RPA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = form_rpa_b_matrix_mo_singlet_full(tei_mo, nocc_alph)
                elif hamiltonian == Hamiltonian.RPA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = form_rpa_b_matrix_mo_triplet_full(tei_mo, nocc_alph)
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = np.zeros(shape=(nov_alph, nov_alph))
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_full(self.moenergies[0], tei_mo, nocc_alph)
                    B = np.zeros(shape=(nov_alph, nov_alph))
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                if hamiltonian == Hamiltonian.RPA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_partial(
                        self.moenergies[0], tei_mo_ovov, tei_mo_oovv
                    )
                    B = form_rpa_b_matrix_mo_singlet_partial(tei_mo_ovov)
                elif hamiltonian == Hamiltonian.RPA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_partial(self.moenergies[0], tei_mo_oovv)
                    B = form_rpa_b_matrix_mo_triplet_partial(tei_mo_ovov)
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_partial(
                        self.moenergies[0], tei_mo_ovov, tei_mo_oovv
                    )
                    B = np.zeros(shape=(nov_alph, nov_alph))
                elif hamiltonian == Hamiltonian.TDA and spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_partial(self.moenergies[0], tei_mo_oovv)
                    B = np.zeros(shape=(nov_alph, nov_alph))

            G = np.block([[A, B], [-B, -A]])
            self.explicit_hessian = G

        else:
            # TODO UHF
            pass

    def diagonalize_explicit_hessian(self) -> None:
        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.occupations
        nov_alph = nocc_alph * nvirt_alph
        nov_beta = nocc_beta * nvirt_beta  # noqa: F841
        if not self.is_uhf:
            eigvals, eigvecs = sp.linalg.eig(self.explicit_hessian)
            # Sort from lowest to highest eigenvalue (excitation
            # energy).
            idx = eigvals.argsort()
            self.eigvals = eigvals[idx]
            # Each eigenvector is a column vector.
            self.eigvecs = eigvecs[:, idx]
            # Fix the ordering of everything. The first eigenvectors
            # are those with negative excitation energies.
            self.eigvals = self.eigvals[nov_alph:]
            self.eigvecs = self.eigvecs[:, nov_alph:]
        else:
            # TODO UHF
            pass

    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(program, Program)
        # TODO program_obj
        if not self.tei_mo:
            self.form_tei_mo(program, program_obj, AO2MOTransformationType.partial)
        self.form_explicit_hessian(hamiltonian, spin, None)
        self.diagonalize_explicit_hessian()


class ExactDiagonalizationSolverTDA(ExactDiagonalizationSolver, EigSolverTDA):
    """Get the eigenvectors and eigenvalues of the TDA Hamiltonian by forming the
    electronic Hessian explicitly and diagonalizing it exactly.
    """

    def __init__(
        self, mocoeffs: np.ndarray, moenergies: np.ndarray, occupations: np.ndarray
    ) -> None:
        super().__init__(mocoeffs, moenergies, occupations)

    def form_explicit_hessian(self, hamiltonian: Hamiltonian, spin: Spin, frequency: float) -> None:
        assert hasattr(self, "tei_mo")
        assert self.tei_mo is not None
        assert len(self.tei_mo) in (1, 2, 4, 6)
        assert isinstance(self.tei_mo_type, AO2MOTransformationType)

        assert hamiltonian == Hamiltonian.TDA
        assert isinstance(spin, Spin)
        assert isinstance(frequency, (float, type(None)))

        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.occupations
        nov_alph = nocc_alph * nvirt_alph  # noqa: F841
        nov_beta = nocc_beta * nvirt_beta  # noqa: F841

        if not self.is_uhf:
            # Set up "function pointers".
            if self.tei_mo_type == AO2MOTransformationType.full:
                assert len(self.tei_mo) == 1
                tei_mo = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                assert len(self.tei_mo) == 2
                tei_mo_ovov = self.tei_mo[0]  # ty: ignore[index-out-of-bounds]
                tei_mo_oovv = self.tei_mo[1]  # ty: ignore[index-out-of-bounds]

            if self.tei_mo_type == AO2MOTransformationType.full:
                if spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_full(self.moenergies[0], tei_mo, nocc_alph)
                elif spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_full(self.moenergies[0], tei_mo, nocc_alph)
            elif self.tei_mo_type == AO2MOTransformationType.partial:
                if spin == Spin.singlet:
                    A = form_rpa_a_matrix_mo_singlet_partial(
                        self.moenergies[0], tei_mo_ovov, tei_mo_oovv
                    )
                elif spin == Spin.triplet:
                    A = form_rpa_a_matrix_mo_triplet_partial(self.moenergies[0], tei_mo_oovv)

            self.explicit_hessian = A

        else:
            # TODO UHF
            pass

    def diagonalize_explicit_hessian(self) -> None:
        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.occupations
        nov_alph = nocc_alph * nvirt_alph  # noqa: F841
        nov_beta = nocc_beta * nvirt_beta  # noqa: F841
        if not self.is_uhf:
            eigvals, eigvecs = sp.linalg.eig(self.explicit_hessian)
            # Sort from lowest to highest eigenvalue (excitation
            # energy).
            idx = eigvals.argsort()
            self.eigvals = eigvals[idx]
            # Each eigenvector is a column vector.
            self.eigvecs = eigvecs[:, idx]
        else:
            # TODO UHF
            raise NotImplementedError



import numpy as np

from pymolresponse.interfaces.pyscf import integrals, molecules


def test_integrals_pyscf() -> None:
    mol = molecules.molecule_water_sto3g()
    mol.build()
    integral_generator = integrals.IntegralsPyscf(mol)
    np.testing.assert_equal(
        mol.intor("cint1e_r_sph", comp=3), integral_generator.integrals(integrals.DIPOLE)
    )
    np.testing.assert_equal(
        mol.intor("cint1e_cg_irxp_sph", comp=3),
        integral_generator.integrals(integrals.ANGMOM_COMMON_GAUGE),
    )


def test_jk_pyscf() -> None:
    mol = molecules.molecule_water_sto3g()
    mol.build()
    jk_generator = integrals.JKPyscf(mol)  # noqa: F841

    # print(jk_generator.compute_from_mocoeffs())



from pathlib import Path
from typing import Union

from pymolresponse.data import REFDIR
from pymolresponse.interfaces.dalton.utils import dalton_label_to_operator
from pymolresponse.operators import Operator
from pymolresponse.tests.test_calculators import calculate_disk_rhf


def run_dalton_label_to_operator(
    dalton_label: str,
    operator_label: str,
    slice_idx: int,
    is_imaginary: bool,
    is_spin_dependent: bool,
) -> Operator:
    operator = dalton_label_to_operator(dalton_label)
    assert operator.label == operator_label
    assert operator.slice_idx == slice_idx
    assert operator.is_imaginary == is_imaginary
    assert operator.is_spin_dependent == is_spin_dependent
    return operator


# def run_reference_disk_rhf(testcase):


def run_as_many_tests_as_possible_rhf_disk(testcase: Union[Path, str]) -> None:
    testcasedir = REFDIR / testcase

    thresh = 5.0e-3

    # These are operators that we can't run for some reason.
    exclude_parts = ("_spnorb",)

    entries = []

    with open(testcasedir / "ref") as fh:
        for line in fh:
            tokens = line.split()
            assert len(tokens) == 6
            hamiltonian = tokens[0]
            spin = tokens[1]
            frequency = tokens[2]
            label_1 = tokens[3]
            label_2 = tokens[4]
            ref = float(tokens[5])

            entry = (hamiltonian, spin, frequency, label_1, label_2, ref)
            entries.append(entry)

    for entry in entries:
        assert len(entry) == 6
        hamiltonian, spin, frequency, label_1, label_2, ref = entry
        ignore_label_1 = any(exclude_part in label_1 for exclude_part in exclude_parts)
        ignore_label_2 = any(exclude_part in label_2 for exclude_part in exclude_parts)
        if not ignore_label_1 and not ignore_label_2:
            res = calculate_disk_rhf(testcasedir, hamiltonian, spin, frequency, label_1, label_2)
            diff = abs(ref) - abs(res)
            format_list = (testcase, hamiltonian, spin, label_1, label_2, ref, res, diff)
            print("{} {} {} {:10} {:10} {:+10e} {:+10e} {:+10e}".format(*format_list))
            assert diff < thresh


def run_as_many_tests_as_possible_uhf_disk(testcase: Union[Path, str]) -> None:
    from .test_calculators import calculate_disk_uhf

    testcasedir = REFDIR / testcase

    thresh = 1.0e-1

    # These are operators that we can't run for some reason.
    exclude_parts = ("_spnorb",)

    entries = []

    with open(testcasedir / "ref") as fh:
        for line in fh:
            tokens = line.split()
            assert len(tokens) == 6
            hamiltonian = tokens[0]
            spin = tokens[1]
            frequency = tokens[2]
            label_1 = tokens[3]
            label_2 = tokens[4]
            ref = float(tokens[5])

            entry = (hamiltonian, spin, frequency, label_1, label_2, ref)
            entries.append(entry)

    for entry in entries:
        assert len(entry) == 6
        hamiltonian, spin, frequency, label_1, label_2, ref = entry
        ignore_label_1 = any(exclude_part in label_1 for exclude_part in exclude_parts)
        ignore_label_2 = any(exclude_part in label_2 for exclude_part in exclude_parts)
        if not ignore_label_1 and not ignore_label_2:
            res = calculate_disk_uhf(testcasedir, hamiltonian, spin, frequency, label_1, label_2)
            diff = abs(ref) - abs(res)
            format_list = (testcase, hamiltonian, spin, label_1, label_2, ref, res, diff)
            print("{} {} {} {:10} {:10} {:+10e} {:+10e} {:+10e}".format(*format_list))
            assert diff < thresh



from pymolresponse.operators import Operator


def clean_dalton_label(original_label: str) -> str:
    """Operator/integral labels in DALTON are in uppercase and may have
    spaces in them; replace spaces with underscores and make all
    letters lowercase.

    >>> clean_dalton_label("PSO 002")
    'pso_002'
    """
    return original_label.lower().replace(" ", "_")


def dalton_label_to_operator(label: str) -> Operator:
    label = clean_dalton_label(label)

    coord1_to_slice = {"x": 0, "y": 1, "z": 2}
    coord2_to_slice = {
        "xx": 0,
        "xy": 1,
        "xz": 2,
        "yy": 3,
        "yz": 4,
        "zz": 5,
        "yx": 1,
        "zx": 2,
        "zy": 4,
    }
    slice_to_coord1 = {v: k for (k, v) in coord1_to_slice.items()}

    # dipole length
    if "diplen" in label:
        operator_label = "dipole"
        _coord = label[0]
        slice_idx = coord1_to_slice[_coord]
        is_imaginary = False
        is_spin_dependent = False
    # dipole velocity
    elif "dipvel" in label:
        operator_label = "dipvel"
        _coord = label[0]
        slice_idx = coord1_to_slice[_coord]
        is_imaginary = True
        is_spin_dependent = False
    # angular momentum
    elif "angmom" in label:
        operator_label = "angmom"
        _coord = label[0]
        slice_idx = coord1_to_slice[_coord]
        is_imaginary = True
        is_spin_dependent = False
    # spin-orbit
    elif "spnorb" in label:
        operator_label = "spinorb"
        _coord = label[0]
        slice_idx = coord1_to_slice[_coord]
        is_imaginary = True
        is_spin_dependent = True
        _nelec = label[1]
        if _nelec in ("1", "2"):
            operator_label += _nelec
        # combined one- and two-electron
        elif _nelec in (" ", "_"):
            operator_label += "c"
        else:
            pass
    # Fermi contact
    elif "fc" in label:
        operator_label = "fermi"
        _atomid = label[6 : 6 + 2]
        slice_idx = int(_atomid) - 1
        is_imaginary = False
        is_spin_dependent = True
    # spin-dipole
    elif "sd" in label:
        operator_label = "sd"
        _coord_atom = label[3 : 3 + 3]
        _coord = label[7]
        _atomid = (int(_coord_atom) - 1) // 3
        _coord_1 = (int(_coord_atom) - 1) % 3
        _coord_2 = slice_to_coord1[_coord_1] + _coord
        slice_idx = (6 * _atomid) + coord2_to_slice[_coord_2]
        is_imaginary = False
        is_spin_dependent = True
    # TODO SD+FC?
    # nucleus-orbit
    elif "pso" in label:
        operator_label = "pso"
        # TODO coord manipulation
        is_imaginary = True
        # TODO is this correct?
        is_spin_dependent = False
        # FIXME
        slice_idx = None
    else:
        operator_label = ""
        is_imaginary = None
        is_spin_dependent = None
        slice_idx = None

    operator = Operator(
        label=operator_label,
        is_imaginary=is_imaginary,
        is_spin_dependent=is_spin_dependent,
        slice_idx=slice_idx,
    )

    return operator



"""Wrapper for performing a dipole polarizability calculation."""

from typing import Sequence

import numpy as np

from pymolresponse.core import Program
from pymolresponse.cphf import CPHF
from pymolresponse.molecular_property import ResponseProperty
from pymolresponse.operators import Operator


class Polarizability(ResponseProperty):
    """Wrapper for performing a dipole polarizability calculation."""

    def __init__(
        self,
        program: Program,
        program_obj,
        driver: CPHF,
        mocoeffs: np.ndarray,
        moenergies: np.ndarray,
        occupations: np.ndarray,
        *,
        frequencies: Sequence[float] = [0.0],
    ) -> None:
        super().__init__(
            program, program_obj, driver, mocoeffs, moenergies, occupations, frequencies=frequencies
        )
        self.polarizabilities = []

    def form_operators(self) -> None:
        if self.program == Program.PySCF:
            from pymolresponse.interfaces.pyscf import integrals

            integral_generator = integrals.IntegralsPyscf(self.program_obj)
        elif self.program == Program.Psi4:
            from pymolresponse.interfaces.psi4 import integrals

            integral_generator = integrals.IntegralsPsi4(self.program_obj)
        else:
            raise RuntimeError

        operator_diplen = Operator(
            label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
        )
        operator_diplen.ao_integrals = integral_generator.integrals(integrals.DIPOLE)
        self.driver.add_operator(operator_diplen)

    def form_results(self) -> None:
        assert len(self.driver.results) == len(self.frequencies)

        for idxf, frequency in enumerate(self.frequencies):
            results = self.driver.results[idxf]
            assert results.shape == (3, 3)
            self.polarizabilities.append(results)
